{$F+}
Unit VGAGUI;

Interface

Uses Crt,Dos,Objects,Drivers,Strings;

Const
  _32BIT_OPERAND = $66;
  VGA_SEGMENT    = $A000;
  SC_INDEX       = $3C4;   { Sequence Controller Index register   }
  GC_INDEX       = $3CE;   { Graphics Controller Index register   }
  CRTC_INDEX     = $3D4;   { CRT Controller Index register        }
  MISC_OUTPUT    = $3C2;   { Miscellaneous Output Register        }
  MISC_INPUT     = $3CC;   { Miscellaneous Input Register         }
  MISCELLANEOUS  = 6;      { Moscellaneous register index in GC   }
  MEMORY_MODE    = 4;      { Memory Mode register index in SC     }
  GRAPHICS_MODE  = 5;      { Graphics Mode register index in GC   }
  MAX_SCAN_LINE  = 9;      { Maximum Scan Line reg index in CRTC  }
  UNDERLINE      = $14;    { Underline Location reg index in CRTC }
  MODE_CONTROL   = $17;    { Mode Control register index in CRTC  }
  MAP_MASK       = 2;      { Map Mask register index in SC        }

  HORZ_TOTAL         = 0;
  HORZ_DISPLAY_END   = 1;
  START_HORZ_BLANK   = 2;
  END_HORZ_BLANK     = 3;
  START_HORZ_RETRACE = 4;
  END_HORZ_RETRACE   = 5;
  VERT_TOTAL         = 6;
  OVERFLOW           = 7;
  START_VERT_RETRACE = $10;
  END_VERT_RETRACE   = $11;
  VERT_DISPLAY_END   = $12;
  CRTC_OFFSET        = $13;
  START_VERT_BLANK   = $15;
  END_VERT_BLANK     = $16;
  LINE_COMPARE       = $18;

Type
  PStringObj = ^TStringObj;
  TStringObj = Object(TObject)
    P : PChar;
    Constructor Init(St: String);
    Destructor  Done; Virtual;
  End;

  TVGACompression = (vcNone,vcRLE,vcLempelZiv);

  TVGAColorName = (clBlack,       clMaroon,       clGreen,        clOlive,
                   clNavy,        clPurple,       clTeal,         clGray,
                   clSilver,      clRed,          clLime,         clYellow,
                   clBlue,        clFuschia,      clAqua,         clWhite,
                   cl3DDkShadow,  cl3DLight,      clActiveBorder, clActiveCaption,
                   clBtnFace,     clBtnHighlight, clBtnShadow,    clBtnText,
                   clCaptionText, clGrayText,     clHighlight,    clHighlightText,
                   clMenu,        clMenuText,     clScrollBar,    clWindow,
                   clWindowFrame, clWindowText,   clNoHighlight,  clNoHighlightText);

  TEventType = (etOnMouseUp,etOnMouseDown,etOnMouseMove,etOnKeyDown,etOnClick,
                etOnShow,etOnClose,etOnPaint,etOnChange,etOnDblClick,etOnRun);

  TOrientation = (orHorizontal,orVertical);

  TMsgBoxButton = Set Of (mbOK,mbCancel);
  TMsgBoxResult = (idOK,idCancel);

  PSortedStringCollection = ^TSortedStringCollection;
  TSortedStringCollection = Object(TSortedCollection)
    Function Compare(Key1,Key2: Pointer): Integer; Virtual;
    Function KeyOf(Item: Pointer): Pointer;        Virtual;
  End;

  PVGAControl = ^TVGAControl;
  TNotifyEvent = Procedure(Sender: PVGAControl; Var Event: TEvent);
  TVGAControl = Object(TObject)
  Private
    Focused  : Boolean;
    Visible  : Boolean;
    FOnPaint : TNotifyEvent;
    Procedure   Paint; Virtual;
    Function    GetAbsLeft: Integer; Virtual;
    Function    GetAbsTop: Integer;  Virtual;
    Procedure   SetVisibleControl(C: PVGAControl; V: Boolean); Virtual;
    Procedure   SetFocusedControl(C: PVGAControl; V: Boolean); Virtual;
    Procedure   HandleEvent(Var Event: TEvent);                Virtual;
    Procedure   MouseOff(Var L,T: Integer);                    Virtual;
  Public
    Left     : Integer;
    Top      : Integer;
    Width    : Integer;
    Height   : Integer;
    Enabled  : Boolean;
    Parent   : PVGAControl;
    Owner    : PVGAControl;
    Color    : TVGAColorName;
    TabOrder : Integer;         { -1 if not focusable }
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer;
                     AVisible,AEnabled: Boolean; AColor: TVGAColorName);
    Destructor  Done; Virtual;
    Function    GetVisible: Boolean;
    Procedure   SetVisible(V: Boolean);
    Function    GetFocused: Boolean;
    Procedure   SetFocused(V: Boolean);
    Function    GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl; Virtual;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent);   Virtual;
    Procedure   AddTo(AParent,AOwner: PVGAControl);                    Virtual;
    Function    InBounds(X,Y: Integer): Boolean;                       Virtual;
  End;

  PVGAlabel = ^TVGALabel;
  TVGALabel = Object(TVGAControl)
  Private
    Text : String;
    Procedure   Paint; Virtual;
  Public
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer;
                     AVisible,AEnabled: Boolean; AColor: TVGAColorName; ACaption: String);
    Destructor  Done; Virtual;
    Function    GetText: String;
    Procedure   SetText(St: String);
  End;

  PVGAComponent = ^TVGAComponent;
  TVGAComponent = Object(TVGAControl)
  Private
    FOnMouseDown : TNotifyEvent;
    FOnMouseUp   : TNotifyEvent;
    FOnMouseMove : TNotifyEvent;
    FOnKeyDown   : TNotifyEvent;
    Procedure   HandleEvent(Var Event: TEvent); Virtual;
  Public
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer;
                     AVisible,AEnabled: Boolean; AColor: TVGAColorName);
    Destructor  Done; Virtual;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent); Virtual;
  End;

  PVGAEdit = ^TVGAEdit;
  TVGAEdit = Object(TVGAComponent)
  Private
    TextTop : Integer;
    Cursor  : Integer;
    Text    : String;
    Select  : Integer;
    Procedure   Paint; Virtual;
    Procedure   HandleEvent(Var Event: TEvent); Virtual;
  Public
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                     AColor: TVGAColorName; Const AText: String);
    Destructor  Done; Virtual;
    Procedure   SetText(St: String);
    Function    GetText: String;
  End;

  PVGAButton = ^TVGAButton;
  TVGAButton = Object(TVGAComponent)
  Private
    FOnClick : TNotifyEvent;
    Procedure   Paint; Virtual;
    Procedure   HandleEvent(Var Event: TEvent); Virtual;
  Public
    Caption : String;
    Pressed : Boolean;
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                     AColor: TVGAColorName; Const ACaption: String);
    Destructor  Done; Virtual;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent); Virtual;
  End;

  PVGASpeedButton = ^TVGASpeedButton;
  TVGASpeedButton = Object(TVGAButton)
  Private
    Procedure   Paint; Virtual;
    Procedure   HandleEvent(Var Event: TEvent); Virtual;
  Public
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                     AColor: TVGAColorName);
    Destructor  Done; Virtual;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent); Virtual;
  End;

  PVGAScrollBar = ^TVGAScrollBar;
  TVGAScrollBar = Object(TVGAComponent)
  Private
    Orientation : TOrientation;
    FOnChange   : TNotifyEvent;
    btnUp       : PVGASpeedButton;
    btnDown     : PVGASpeedButton;
    Procedure   Paint; Virtual;
    Procedure   HandleEvent(Var Event: TEvent); Virtual;
    Procedure   btnUpClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure   btnDownClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure   btnUpPaint(Sender: PVGAControl; Var Event: TEvent);
    Procedure   btnDownPaint(Sender: PVGAControl; Var Event: TEvent);
  Public
    Min         : LongInt;
    Max         : LongInt;
    Position    : LongInt;
    LargeChange : LongInt;
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                     AColor: TVGAColorName; AType: TOrientation; AMin,AMax,APosition,ALargeChange: LongInt);
    Destructor  Done;                                                  Virtual;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent);   Virtual;
    Procedure   AddTo(AParent,AOwner: PVGAControl);                    Virtual;
    Function    GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl; Virtual;
  End;

  PVGAListBox = ^TVGAListBox;
  TVGAListBox = Object(TVGAComponent)
  Private
    Orientation : TOrientation;
    sbSelect    : PVGAScrollBar;
    FOnChange   : TNotifyEvent;
    FOnDblClick : TNotifyEvent;
    Procedure   Paint; Virtual;
    Procedure   HandleEvent(Var Event: TEvent); Virtual;
    Procedure   sbSelectChange(Sender: PVGAControl; Var Event: TEvent);
  Public
    ColWidth  : Integer;
    RowHeight : Integer;
    ItemIndex : Integer;
    Items     : PCollection;
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                     AColor: TVGAColorName; AType: TOrientation; AColWidth,ARowHeight: Integer);
    Destructor  Done;                                                  Virtual;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent);   Virtual;
    Procedure   AddTo(AParent,AOwner: PVGAControl);                    Virtual;
    Function    GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl; Virtual;
    Procedure   AddString(St: String);
    Procedure   Clear;
  End;

  TBorderStyle = (bsNone,bsSingle,bsDialog);
  PVGAForm = ^TVGAForm;
  TVGAForm = Object(TVGAComponent)
  Private
    TabNum           : Integer;
    MaxTab           : Integer;
    FocusedComponent : Integer;
    Modal            : Boolean;
    Updating         : Boolean;
    FOnShow          : TNotifyEvent;
    FOnClose         : TNotifyEvent;
    Procedure   Paint;                                         Virtual;
    Function    GetAbsLeft: Integer;                           Virtual;
    Function    GetAbsTop: Integer;                            Virtual;
    Procedure   FocusNextComponent;
    Procedure   SetVisibleControl(C: PVGAControl; V: Boolean); Virtual;
    Procedure   SetFocusedControl(C: PVGAControl; V: Boolean); Virtual;
    Procedure   HandleEvent(Var Event: TEvent);                Virtual;
  Public
    Caption        : String;
    Controls       : PCollection;
    BorderStyle    : TBorderStyle;
    ZOrder         : Integer;
    Default        : PVGAControl;
    DisposeOnClose : Boolean;
    Constructor Init(ALeft,ATop,AWidth,AHeight: Integer; AEnabled: Boolean;
                     AColor: TVGAColorName; Const ACaption: String; ABorder: TBorderStyle);
    Destructor  Done;                                                  Virtual;
    Function    ClientWidth: Integer;
    Function    ClientHeight: Integer;
    Procedure   AddControl(Control: PVGAControl);
    Procedure   Show;
    Procedure   ShowModal;
    Procedure   Close;
    Function    GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl; Virtual;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent);   Virtual;
  End;

  PVGAChooser = ^TVGAChooser;
  TVGAChooser = Object(TVGAForm)
  Private
    btnOK       : PVGAButton;
    btnCancel   : PVGAButton;
    edtFile     : PVGAEdit;
    lbMain      : PVGAListBox;
    lblFileName : PVGALabel;
    lblPath     : PVGALabel;
    Filter      : String;
    InitialDir  : String;
    Procedure  btnOkClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure  btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure  lbMainChange(Sender: PVGAControl; Var Event: TEvent);
    Procedure  lbMainDblClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure  LoadListBox;
    Procedure  HandleEvent(Var Event: TEvent); Virtual;
  Public
    FileName   : String;
    Chose      : Boolean;
    Constructor Init(AEnabled: Boolean; Const ACaption,AFilter,AInitialDir: String);
    Destructor  Done; Virtual;
    Procedure   SetFilter(Const AFilter: String);
    Function    GetFilter: String;
    Procedure   SetInitialDir(Const AInitialDir: String);
    Function    GetInitialDir: String;
  End;

  PVGAMessageBox = ^TVGAMessageBox;
  TVGAMessageBox = Object(TVGAForm)
  Private
    btnOK      : PVGAButton;
    btnCancel  : PVGAButton;
    lblMessage : PVGALabel;
    Procedure  btnOkClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure  btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure  HandleEvent(Var Event: TEvent); Virtual;
  Public
    Result : TMSGBoxResult;
    Constructor Init(AEnabled: Boolean; Const ACaption,AMessage: String; Buttons: TMSGBoxButton);
    Destructor  Done; Virtual;
  End;

  PVGAInputBox = ^TVGAInputBox;
  TVGAInputBox = Object(TVGAForm)
  Private
    btnOK      : PVGAButton;
    btnCancel  : PVGAButton;
    lblMessage : PVGALabel;
    edtInput   : PVGAEdit;
    Text       : String;
    Procedure  btnOkClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure  btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
    Procedure  HandleEvent(Var Event: TEvent); Virtual;
  Public
    InitialText : String;
    Constructor Init(AEnabled: Boolean; Const ACaption,AMessage,AInitialText: String);
    Destructor  Done; Virtual;
    Function    GetText: String;
  End;

  PWindowBackgroundData = ^TWindowBackgroundData;
  TWindowBackgroundData = Object(TObject)
    Data  : Pointer;
    Size  : Word;
    Lines : Word;
    Constructor Init(ASize,ALines: Word);
    Destructor  Done; Virtual;
  End;

  PWindowBackground = ^TWindowBackground;
  TWindowBackground = Object(TObject)
    X1,Y1,X2,Y2 : Integer;
    Data        : PCollection;
    Size        : LongInt;
    Compression : TVGACompression;
    Constructor Init(AX1,AY1,AX2,AY2: Integer; AComp: TVGACompression);
    Destructor  Done; Virtual;
    Procedure   Allocate(ASize: LongInt);
  End;

  PVGAApplication = ^TVGAApplication;
  TVGAApplication = Object(TObject)
  Private
    Stack       : PCollection;
    Forms       : PCollection;
    ZOrder      : Integer;
    FOnRun      : TNotifyEvent;
    SaveBG      : Boolean;
    Running     : Boolean;
    MouseWasOn  : Boolean;
  Public
    Constructor Init(AWidth,AHeight: Integer; AModeX,ASaveBG,LoadPal: Boolean);
    Destructor  Done;                                                Virtual;
    Procedure   AddForm(Form: PVGAForm);
    Procedure   Run;
    Procedure   Idle;                                                Virtual;
    Procedure   ShowForm(Form: PVGAForm);
    Procedure   CloseForm(Form: PVGAForm);
    Procedure   SaveBackground(W: PWindowBackground);                Virtual;
    Procedure   RestoreBackground(W: PWindowBackground);             Virtual;
    Procedure   HandleEvent(Var Event: TEvent);
    Procedure   Save(X1,Y1,X2,Y2: Integer);
    Procedure   Restore;
    Procedure   SortForms;
    Procedure   ProcessMessages;
    Function    ShowMessage(Const ACaption,AMessage: String; Buttons: TMsgBoxButton): TMsgBoxResult;
    Procedure   PaintForms;
    Procedure   AssignEvent(E: TEventType; EventProc: TNotifyEvent); Virtual;
    Function    InputBox(Const ACaption,AMessage,AInitialText: String): String;
    Function    FileDialog(Const ACaption,AFilter,AInitialDir: String): String;
    Procedure   VGAInitEvents;
    Procedure   VGADoneEvents;
  End;

Const
  ScrollBarSize  : Integer         = 9;
  ModeX          : Boolean         = False;
  VGAApplication : PVGAApplication = Nil;

Var
  ScreenWidth    : Integer;
  ScreenHeight   : Integer;

Procedure PutSmChar(X,Y: Word; C: Char; Color: Word);
Procedure WriteSmOut(X,Y: Word; St: String; Color: Word);
Procedure PutPixel(X,Y: Integer; Color: Byte);
Procedure Box(X1,Y1,X2,Y2: Integer; Color: Byte);
Procedure FBox(X1,Y1,X2,Y2: Integer; Color: Byte);
Procedure GetImage(X1,Y1,X2,Y2: Word; Var P);
Procedure PutImage(X1,Y1,W,H: Word; Var P);
Procedure PutMaskedImage(X1,Y1,W,H,W1,H1: Word; Var Img,Mask);
Function  FileExists(FileName: String): Boolean;
Function  ValidFileName(FileName: String): Boolean;
Function  IsDirectory(FileName: String): Boolean;
Procedure InitGUI(AWidth,AHeight: Integer; AModeX,LoadPal: Boolean);

Implementation

Uses SmalFont,SmModeX,Keyboard,MouseX;

Type
  TRGBA = Packed Record
    R,G,B,A : Byte;
  End;
  TVGAColor = LongInt;

Const
  BorderWidth   = 2;
  CaptionHeight = 12;
  VGAColorIdeal   : Array[TVGAColorName] Of TVGAColor =
   ($00000000, $00000080, $00008000, $00008080,
    $00800000, $00800080, $00808000, $00808080,
    $00D2D2D2, $000000FF, $0000FF00, $0000FFFF,
    $00FF0000, $00FF00FF, $00FFFF00, $00FFFFFF,
    $00000000, $00FFFFFF, $00D2D2D2, $00FF0000,
    $00D2D2D2, $00FFFFFF, $00808080, $00000000,
    $00FFFFFF, $00000000, $006A240A, $00FFFFFF,
    $00D2D2D2, $00000000, $00E0E0E0, $00FFFFFF,
    $00D2D2D2, $00000000, $00A0A0A0, $00000000);
  VGAColorDef     : Array[TVGAColorName] Of TVGAColor =
   ($01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000,
    $01000000, $01000000, $01000000, $01000000);

Var
  VGAColors    : Array[TVGAColorName] Of Byte;
  VGAColorDist : Array[TVGAColorName] Of Single;
  DummyEvent   : TEvent;

Procedure VGAShowMouse;
Begin
  If (ScreenWidth <> 320) Or ModeX Then _MouseOn Else ShowMouse;
End; { VGAShowMouse }

Procedure VGAHideMouse;
Begin
  If (ScreenWidth <> 320) Or ModeX Then _MouseOff Else HideMouse;
End; { VGAHideMouse }

Function FileExists(FileName: String): Boolean;
Var Fil: File;
Begin
  Assign(Fil,FileName);
  {$I-}
  Reset(Fil);
  {$I+}
  If (IOResult = 0) Then
  Begin
    Close(Fil);
    FileExists := True;
  End
  Else FileExists := False;
End; { FileExists }

Function ValidFileName(FileName: String): Boolean;
Var
  B : Boolean;
  F : File;

Begin
  B := FileExists(FileName);
  If Not B Then
  Begin
    Assign(F,FileName);
    {$I-}
    ReWrite(F);
    {$I+}
    If (IOResult = 0) Then
    Begin
      Close(F);
      Erase(F);
      B := True;
    End;
  End;
  ValidFileName := B;
End; { ValidFileName }

Function IsDirectory(FileName: String): Boolean;
Var S: SearchRec;
Begin
  If (Pos('?',FileName) <> 0) Or
     (Pos('*',FileName) <> 0) Then IsDirectory := False
  Else
  Begin
    FindFirst(FileName,AnyFile,S);
    IsDirectory := (DOSError = 0) And ((S.Attr And Directory) <> 0);
  End;
End; { IsDirectory }

{ -------------------------- GUI Initialization -------------------------- }

Procedure CoordFixup; Far; Assembler;
Asm
  MOV   AX,DX
  SUB   DX,DX
  DIV   WORD PTR HeightMul
  MOV   CX,AX
  MOV   AX,SI
  SUB   DX,DX
  DIV   WORD PTR WidthMul
  MOV   SI,AX
  MOV   DX,CX
End; { CoordFixup }

Procedure ModifyCode;
Type
  BPtr = ^Byte;
  WPtr = ^Word;
  LPtr = ^LongInt;

Var
  I     : Word;
  P1,P2 : LPtr;
  P3    : WPtr;
  P4    : BPtr;
  Found : Boolean;

Begin
  { ----------------------------------------------------------------------- }
  { Modify the code in the mouse handler from Turbo Vision DRIVERS.PAS to   }
  { NOT merely shift the vertical and horizontal positions right by 3 bits, }
  { since we are operating in graphics mode.  Instead we will call a        }
  { subroutine which will divide by the appropriate value.                  }
  {                                                                         }
  { This depends on a DRIVERS.PAS that has NOT been modified!!!!            }
  {                                                                         }
  { -----------------------------------+----------------------------------- }
  { OLD CODE                           | NEW CODE                           }
  { -----------------------------------+----------------------------------- }
  { 8E DE   MOV   DS,SI                | <no change>                        }
  { 89 CE   MOV   SI,CX                | <no change>                        }
  { B1 03   MOV   CL,3                 | 9A ofs seg   CALL FAR CoordFixup   }
  { D3 EE   SHR   SI,CL   ; X value    | 90           NOP                   }
  { D3 EA   SHR   DX,CL   ; Y value    |                                    }
  { -----------------------------------+----------------------------------- }

  P1    := @InitEvents;  { Search the same code segment as GetMouseEvent }
  I := Ofs(P1^);
  If I >= 512 Then Dec(I,512) Else I := 0;
  P1    := Ptr(Seg(P1^),I);
  P2    := Pointer(LongInt(P1) + 4);
  P3    := Pointer(LongInt(P2) + 4);
  I     := 512;             { Search up to this many bytes }
  Found := False;
  While (I > 0) And Not Found Do
  Begin
    If (P1^ = $CE89DE8E) And   { Replace last 6 bytes }
       (P2^ = $EED303B1) And
       (P3^ = $EAD3)     Then
    Begin
      P4 := BPtr(P2);
      P4^ := $9A;           { Far call to absolute address }
      Inc(LongInt(P2));
      P2^ := LongInt(@CoordFixup);
      P4  := BPtr(P2);
      Inc(LongInt(P4),4);
      P4^ := $90;           { NOP }
      Found := True;
    End
    Else
    Begin
      Dec(I);
      Inc(LongInt(P1));
      Inc(LongInt(P2));
      Inc(LongInt(P3));
    End;
  End; { While }
End; { ModifyCode }

Procedure InitGUI(AWidth,AHeight: Integer; AModeX,LoadPal: Boolean);
Var
  Pal   : TRGBA;
  I     : Word;
  J     : TVGAColorName;
  R     : Single;

  Function ColorDistance(C1,C2: TRGBA): Single;
  Var
    R1,R2 : Single;
    G1,G2 : Single;
    B1,B2 : Single;

  Begin
    R1 := C1.R / 255;
    G1 := C1.G / 255;
    B1 := C1.B / 255;
    R2 := C2.R / 255;
    G2 := C2.G / 255;
    B2 := C2.B / 255;
    ColorDistance := Sqrt(Sqr(R1 - R2) + Sqr(G1 - G2) + Sqr(B1 - B2));
  End; { ColorDistance }

Begin
  ModeX        := AModeX;
  ScreenWidth  := AWidth;
  ScreenHeight := AHeight;

  { -------------------------------------------------------------------- }
  { Read the graphics palette from the VGA registers and determine which }
  { colors are closest to the predefined RGB values                      }
  { -------------------------------------------------------------------- }

  If LoadPal Then
  Begin

    { Reset the color definitions }

    For J := Low(VGAColorDef) To High(VGAColorDef) Do
     VGAColorDef[J] := $01000000;

    Pal.A      := 0;
    Port[$3C7] := 0;  { Initialize the VGA palette counter }
    For I := 0 To 255 Do
    Begin

      { Get the next color info from the VGA palette registers }

      Pal.R := Port[$3C9] Shl 2;
      Pal.G := Port[$3C9] Shl 2;
      Pal.B := Port[$3C9] Shl 2;

      { Scan through all the predefined colors and see if this color is }
      { closer to any of them than the previous color                   }

      For J := Low(VGAColorDef) To High(VGAColorDef) Do
      Begin
        If TRGBA(VGAColorDef[J]).A <> 0 Then
        Begin
          VGAColors[J]    := I;
          VGAColorDef[J]  := TVGAColor(Pal);
          VGAColorDist[J] := ColorDistance(Pal,TRGBA(VGAColorIdeal[J]));
        End
        Else
        Begin
          R := ColorDistance(Pal,TRGBA(VGAColorIdeal[J]));
          If R < VGAColorDist[J] Then
          Begin
            VGAColors[J]    := I;
            VGAColorDef[J]  := TVGAColor(Pal);
            VGAColorDist[J] := R;
          End;
        End;
      End; { For J }
    End; { For I }
  End;
End; { InitGUI }

{ ------------------------ VGA Graphics Primitves ------------------------ }

Procedure PutSmChar(X,Y: Word; C: Char; Color: Word); Assembler;
Asm
  TEST  BYTE PTR ModeX,0FFh
  JZ    @Normal
  CLI
  MOV   AX,WORD PTR Y
  MOV   DI,WORD PTR ScreenWidth
  SHR   DI,2
  MUL   DI
  MOV   DI,AX
  MOV   AX,WORD PTR X
  SHR   AX,2
  ADD   DI,AX
  MOV   AH,11h
  MOV   CX,WORD PTR X
  AND   CX,3
  ROL   AH,CL
  MOV   AL,MAP_MASK
  MOV   DX,SC_INDEX

  PUSH  DS
  LDS   SI,DWORD PTR FontSmPtr
  MOV   BL,BYTE PTR C
  SUB   BH,BH
  SHL   BX,3
  ADD   SI,BX
  MOV   BX,0A000h
  MOV   ES,BX
  MOV   CX,8
  MOV   BX,WORD PTR Color
@Loop1:
  OUT   DX,AX

  TEST  BYTE PTR [SI],80h
  JNZ   @On0
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Do4
@On0:
  MOV   BYTE PTR [ES:DI],BL
@Do4:
  TEST  BYTE PTR [SI],8
  JNZ   @On4
  MOV   BYTE PTR [ES:DI + 1],BH
  JMP   @Do1
@On4:
  MOV   BYTE PTR [ES:DI + 1],BL
@Do1:

  ROL   AH,1
  ADC   DI,0
  OUT   DX,AX
  TEST  BYTE PTR [SI],40h
  JNZ   @On1
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Do5
@On1:
  MOV   BYTE PTR [ES:DI],BL
@Do5:
  TEST  BYTE PTR [SI],4
  JNZ   @On5
  MOV   BYTE PTR [ES:DI + 1],BH
  JMP   @Do2
@On5:
  MOV   BYTE PTR [ES:DI + 1],BL
@Do2:

  ROL   AH,1
  ADC   DI,0
  OUT   DX,AX
  TEST  BYTE PTR [SI],20h
  JNZ   @On2
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Do3
@On2:
  MOV   BYTE PTR [ES:DI],BL
@Do3:
  ROL   AH,1
  ADC   DI,0
  OUT   DX,AX
  TEST  BYTE PTR [SI],10h
  JNZ   @On3
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Done
@On3:
  MOV   BYTE PTR [ES:DI],BL
@Done:
  ROL   AH,1
  TEST  BYTE PTR X,3
  JZ    @NoDec
  DEC   DI
@NoDec:
  PUSH  AX
  MOV   AX,WORD PTR ScreenWidth
  SHR   AX,2
  ADD   DI,AX
  POP   AX
  INC   SI
  DEC   CX
  JNZ   @Loop1

  POP   DS

  { Restore zero page for reading and writing }

  MOV   DX,03CEh
  MOV   AX,4
  OUT   DX,AX
  MOV   DX,SC_INDEX
  MOV   AX,MAP_MASK
  OUT   DX,AX
  STI
  JMP   @L3

@Normal:
  MOV   AX,WORD PTR Y
  MOV   DI,WORD PTR ScreenWidth
  MUL   DI
  MOV   DI,AX
  ADD   DI,WORD PTR X
  PUSH  DS
  LDS   SI,DWORD PTR FontSmPtr
  MOV   AL,BYTE PTR C
  SUB   AH,AH
  SHL   AX,3
  ADD   SI,AX
  MOV   AX,0A000h
  MOV   ES,AX
  MOV   AL,8
  ADD   DI,5
  MOV   DX,WORD PTR Color
@Loop:
  MOV   BH,BYTE PTR [SI]
  MOV   CX,6
  MOV   BL,4
@L1:
  TEST  BH,BL
  JZ    @L2
  MOV   BYTE PTR [ES:DI],DL
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,WORD PTR ScreenWidth
  ADD   DI,6
  INC   SI
  DEC   AL
  JNZ   @Loop
  POP   DS
  JMP   @L3
@L2:
  MOV   BYTE PTR [ES:DI],DH
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,WORD PTR ScreenWidth
  ADD   DI,6
  INC   SI
  DEC   AL
  JNZ   @Loop
  POP   DS
@L3:
End; { PutSmChar }


Procedure WriteSmOut(X,Y: Word; St: String; Color: Word);
Var I: Byte;
Begin
  If St = '' Then Exit;
  For I := 0 To Length(St) - 1 Do PutSmChar(X + (I * 6),Y,St[I + 1],Color);
End; { WriteSmOut }

Procedure PutPixel(X,Y: Integer; Color: Byte);
Begin
  If ModeX Then
  Begin
    Asm
      CLI
    End; { Asm }
    PutPixelX(X,Y,0,Color);
    Asm
      MOV   DX,03CEh
      MOV   AX,4
      OUT   DX,AX
      MOV   DX,SC_INDEX
      MOV   AX,MAP_MASK
      OUT   DX,AX
      STI
    End;
  End
  Else Mem[SegA000:Y * ScreenWidth + X] := Color;
End; { PutPixel }

Procedure FBox(X1,Y1,X2,Y2: Integer; Color: Byte);
{ ------------------------------------------------------------------------ }
{ Fills the rectangle (X1,Y1,X2,Y2) inclusive with the specified color.    }
{ ------------------------------------------------------------------------ }
Var Addr,Width: Word;
Begin
  If ModeX Then
  Begin
    Asm
      CLI
    End; { Asm }
    FillRectangleX(X1,Y1,X2 + 1,Y2 + 1,0,Color);
    Asm
      MOV   DX,03CEh
      MOV   AX,4
      OUT   DX,AX
      MOV   DX,SC_INDEX
      MOV   AX,MAP_MASK
      OUT   DX,AX
      STI
    End;
  End
  Else
  Begin
    If X1 < 0 Then X1 := 0;
    If Y1 < 0 Then Y1 := 0;
    If X2 >= ScreenWidth  Then X2 := ScreenWidth  - 1;
    If Y2 >= ScreenHeight Then Y2 := ScreenHeight - 1;
    If (X2 < X1) Or (Y2 < Y1) Then Exit;
    Addr  := Y1 * ScreenWidth + X1;
    Width := ScreenWidth;
    Asm
      CLD
      MOV   AX,SegA000
      MOV   ES,AX
      MOV   DI,WORD PTR Addr       { Initial destination address }
      MOV   DX,Y2
      SUB   DX,Y1
      INC   DX                     { DX holds scan line count }
      MOV   BX,X2
      SUB   BX,X1
      INC   BX                     { BX holds the horizontal pixel count }
      MOV   SI,WORD PTR Width
      SUB   SI,BX                  { SI holds the address increment }
      MOV   AL,BYTE PTR Color
      MOV   AH,AL
      MOV   CX,AX
      DB    _32BIT_OPERAND
      SHL   AX,16
      MOV   AX,CX                  { EAX holds four pixels worth of the color }
@MainLoop:
      MOV   CX,BX
      CMP   CX,1
      JNE   @NotVertLine
      MOV   SI,WORD PTR Width
@MainLoopVert:
      MOV   BYTE PTR [ES:DI],AL
      ADD   DI,SI
      DEC   DX
      JNZ   @MainLoopVert
      JMP   @GetOut
@NotVertLine:
      PUSH  BX
      TEST  DI,3                   { See if the destination address is on a }
      JZ    @StartingOnDWORD       { doubleword boundary }
      PUSH  AX
      PUSH  DX
      MOV   AX,4                   { Do 4 - (DI and 3) bytes at most }
      MOV   DX,DI
      AND   DX,3
      SUB   AX,DX
      CMP   AX,CX
      JBE   @DoInitial
      MOV   CX,AX
@DoInitial:
      POP   DX
      POP   AX
      SUB   BX,CX                  { Subtract initial from the total }
      REP   STOSB
      MOV   CX,BX                  { Get the number of remaining pixels }
@StartingOnDWord:
      SHR   CX,2                   { Only write four at a time now }
      JZ    @WriteRemainder
      DB    _32BIT_OPERAND
      REP   STOSW
@WriteRemainder:
      MOV   CX,BX
      AND   CX,3                   { Write any remaining pixels }
      REP   STOSB
      POP   BX
      ADD   DI,SI
      DEC   DX
      JNZ   @MainLoop
@GetOut:
    End; { Asm }
  End;
End; { FBox }

Procedure Box(X1,Y1,X2,Y2: Integer; Color: Byte);
Begin
  If ModeX Then RectangleX(X1,Y1,X2,Y2,0,Color)
  Else
  Begin
    FBox(X1,Y1,X2,Y1,Color);
    FBox(X1,Y1,X1,Y2,Color);
    FBox(X2,Y1,X2,Y2,Color);
    FBox(X1,Y2,X2,Y2,Color);
  End;
End; { Rectangle }

Procedure GetImage(X1,Y1,X2,Y2: Word; Var P);
Var
  X,Y,I,Addr,Dest,X1A,X2A,Addr1 : LongInt;
  P1                            : Pointer;

Begin
  If ModeX Then
  Begin
    Asm
      CLI
    End; { Asm }
    P1 := @P;
    For Y := Y1 To Y2 Do
    Begin
      Addr := (Y * ScreenWidth + X1) Shr 2;
      Dest := (Y - Y1) * (X2 - X1 + 1);
      For I := 0 To 3 Do
      Begin

        { Set the plane read register }

        Asm
          MOV   DX,03CEh
          MOV   AL,4
          MOV   AH,BYTE PTR I
          OUT   DX,AX
        End; { Asm }

        Addr1 := Addr;
        X1A   := X1 Shr 2;
        X2A   := X2 Shr 2;
        If (X1 And 3) > I Then
        Begin
          Inc(X1A);
          Inc(Addr1);
        End;
        If (X2 And 3) < I Then Dec(X2A);
        Asm
          CLD
          MOV   CX,WORD PTR X2A
          SUB   CX,WORD PTR X1A
          JS    @Done
          INC   CX
          PUSH  DS
          MOV   AX,SegA000
          MOV   DS,AX
          MOV   SI,WORD PTR Addr1
          LES   DI,DWORD PTR P
          ADD   DI,WORD PTR Dest
          MOV   AX,WORD PTR I
          MOV   BX,WORD PTR X1
          AND   BX,3
          SUB   AX,BX
          AND   AX,3
          ADD   DI,AX
@L1:
          MOVSB
          ADD   DI,3
          LOOP  @L1
          POP   DS
@Done:
        End; { Asm }
      End;
    End; { For Y }

    { Reset read and write plane registers }

    Asm
      MOV   DX,03CEh
      MOV   AX,4
      OUT   DX,AX
      MOV   DX,SC_INDEX
      MOV   AX,MAP_MASK
      OUT   DX,AX
      STI
    End;
  End
  Else
  Begin
    For Y := Y1 To Y2 Do
    Begin
      Addr := Y * ScreenWidth + X1;
      Dest := (Y - Y1) * (X2 - X1 + 1);
      Move(Mem[SegA000:Addr],Mem[Seg(P):Ofs(P) + Dest],X2 - X1 + 1);
    End; { For Y }
  End;
End; { GetImage }

Procedure PutImage(X1,Y1,W,H: Word; Var P);
Var X,Y,Addr,Source: LongInt;
Begin
  If ModeX Then
  Begin
    Asm
      CLI
    End; { Asm }
    CopySystemToScreenX(0,0,W,H,X1,Y1,@P,0,W,ScreenWidth);
    Asm
      MOV   DX,03CEh
      MOV   AX,4
      OUT   DX,AX
      MOV   DX,SC_INDEX
      MOV   AX,MAP_MASK
      OUT   DX,AX
      STI
    End;
  End
  Else
  Begin
    For Y := Y1 To Y1 + H - 1 Do
    Begin
      Addr   := Y * ScreenWidth + X1;
      Source := (Y - Y1) * W;
      Move(Mem[Seg(P):Ofs(P) + Source],Mem[SegA000:Addr],W);
    End; { For Y }
  End;
End; { PutImage }

Procedure PutMaskedImage(X1,Y1,W,H,W1,H1: Word; Var Img,Mask);
Var
  X,Y,Addr,Source : LongInt;
  P1,P2           : ^Byte;

Begin
  If ModeX Then
  Begin
    Asm
      CLI
    End; { Asm }
    ZapSystemToScreenMaskedX1(0,0,W1,H1,X1,Y1,@Img,0,W,ScreenWidth,@Mask);
    Asm
      MOV   DX,03CEh
      MOV   AX,4
      OUT   DX,AX
      MOV   DX,SC_INDEX
      MOV   AX,MAP_MASK
      OUT   DX,AX
      STI
    End;
  End
  Else
  Begin
    P1 := @Img;
    P2 := @Mask;
    For Y := Y1 To Y1 + H1 - 1 Do
    Begin
      Addr := Y * ScreenWidth + X1;
      For X := X1 To X1 + W1 - 1 Do
      Begin
        If P2^ = 0 Then Mem[SegA000:Addr] := P1^;
        Inc(LongInt(P1));
        Inc(LongInt(P2));
        Inc(Addr);
      End; { For X }
      Inc(LongInt(P1),W - W1);
      Inc(LongInt(P2),W - W1);
    End; { For Y }
  End;
End; { PutMaskedImage }

Procedure Bevel(L,T,Width,Height: Integer; C1,C2,C3,C4: TVGAColorName);
Begin
  FBox(L,T,L + Width - 1,T,VGAColors[C1]);
  FBox(L,T,L,T + Height - 1,VGAColors[C1]);
  FBox(L + Width - 1,T,L + Width - 1,T + Height - 1,VGAColors[C2]);
  FBox(L + 1,T + Height - 1,L + Width - 1,T + Height - 1,VGAColors[C2]);
  FBox(L + 1,T + 1,L + Width - 2,T + 1,VGAColors[C3]);
  FBox(L + 1,T + 1,L + 1,T + Height - 2,VGAColors[C3]);
  FBox(L + Width - 2,T + 2,L + Width - 2,T + Height - 2,VGAColors[C4]);
  FBox(L + 2,T + Height - 2,L + Width - 2,T + Height - 2,VGAColors[C4]);
End; { Bevel }

{ ------------------------------ Class Code ------------------------------ }

{ ----------------------------------------- }
{ TStringObj                                }
{ ----------------------------------------- }

Constructor TStringObj.Init(St: String);
Begin
  Inherited Init;
  GetMem(P,Length(St) + 1);
  StrPCopy(P,St);
End; { TStringObj.Init }

Destructor TStringObj.Done;
Begin
  StrDispose(P);
  Inherited Done;
End; { TStringObj.Done }

{ ----------------------------------------- }
{ TSortedStringCollection                   }
{ ----------------------------------------- }

Function TSortedStringCollection.KeyOf(Item: Pointer): Pointer;
Begin
  KeyOf := PStringObj(Item)^.P;
End; { TSortedStringCollection.KeyOf }

Function TSortedStringCollection.Compare(Key1,Key2: Pointer): Integer;
Begin
  Compare := StrComp(Key1,Key2);
End; { TSortedStringCollection.Compare }

{ ----------------------------------------- }
{ TWindowBackgroundData                     }
{ ----------------------------------------- }

Constructor TWindowBackgroundData.Init(ASize,ALines: Word);
Begin
  Inherited Init;
  GetMem(Data,ASize);
  Size  := ASize;
  Lines := ALines;
End; { TWindowBackgroundData.Init }

Destructor TWindowBackgroundData.Done;
Begin
  FreeMem(Data,Size);
  Inherited Done;
End; { TWindowBackgroundData.Done }

{ ----------------------------------------- }
{ TWindowBackground                         }
{ ----------------------------------------- }

Constructor TWindowBackground.Init(AX1,AY1,AX2,AY2: Integer; AComp: TVGACompression);
Begin
  Inherited Init;
  X1          := AX1;
  Y1          := AY1;
  X2          := AX2;
  Y2          := AY2;
  Data        := New(PCollection,Init(10,10));
  Size        := 0;
  Compression := AComp;
End; { TWindowBackground.Init }

Destructor TWindowBackground.Done;
Begin
  Dispose(Data,Done);
  Inherited Done;
End; { TWindowBackground.Done }

Procedure TWindowBackground.Allocate(ASize: LongInt);
Var
  D : PWindowBackgroundData;
  L : LongInt;
  M : LongInt;
  W : LongInt;

Begin
  Size := ASize;
  W    := X2 - X1 + 1;
  M    := (65520 Div W) * W;
  While ASize > 0 Do
  Begin
    If ASize > M
     Then L := M
     Else L := ASize;
    D := New(PWindowBackgroundData,Init(L,L Div W));
    Data^.Insert(D);
    Dec(ASize,L);
  End; { While }
End; { TWindowBackground.Allocate }

{ ----------------------------------------- }
{ TVGAApplication                           }
{ ----------------------------------------- }

Constructor TVGAApplication.Init(AWidth,AHeight: Integer; AModeX,ASaveBG,LoadPal: Boolean);
Begin
  Inherited Init;
  VGAApplication := @Self;
  InitGUI(AWidth,AHeight,AModeX,LoadPal);
  Stack      := New(PCollection,Init(10,10));
  Forms      := New(PCollection,Init(10,10));
  ZOrder     := -1;
  FOnRun     := Nil;
  SaveBG     := ASaveBG;
  Running    := False;
  MouseWasOn := MouseOn;
End; { TVGAApplication.Init }

Destructor TVGAApplication.Done;
Begin
  Dispose(Stack,Done);
  Dispose(Forms,DOne);
End; { TVGAApplication.Done }

Procedure TVGAApplication.Save(X1,Y1,X2,Y2: Integer);
Var W: PWindowBackground;
Begin
  If SaveBG Then
  Begin
    If X1 < 0 Then X1 := 0;
    If Y1 < 0 Then Y1 := 0;
    If X2 < 0 Then X2 := 0;
    If Y2 < 0 Then Y2 := 0;
    If X1 >= ScreenWidth  Then X1 := ScreenWidth  - 1;
    If Y1 >= ScreenHeight Then Y1 := ScreenHeight - 1;
    If X2 >= ScreenWidth  Then X2 := ScreenWidth  - 1;
    If Y2 >= ScreenHeight Then Y2 := ScreenHeight - 1;
    If (X1 <= X2) And (Y1 <= Y2) Then
    Begin
      W := New(PWindowBackGround,Init(X1,Y1,X2,Y2,vcNone));
      SaveBackground(W);
      Stack^.Insert(W);
    End;
  End;
End; { TVGAApplication.Save }

Procedure TVGAApplication.Restore;
Var W: PWindowBackground;
Begin
  If SaveBG And (Stack^.Count > 0) Then
  Begin
    W := PWindowBackground(Stack^.At(Stack^.Count - 1));
    RestoreBackground(W);
    Stack^.Free(W);
  End
  Else PaintForms;
End; { TVGAApplication.Restore }

Procedure TVGAApplication.SaveBackground(W: PWindowBackground);
Var
  W1,H1,C : LongInt;
  I       : Integer;
  WD      : PWindowBackGroundData;

Begin
  W1 := W^.X2 - W^.X1 + 1;
  H1 := W^.Y2 - W^.Y1 + 1;
  W^.Allocate(W1 * H1);
  C := 0;
  I := 0;
  While I < W^.Data^.Count Do
  Begin
    WD := PWindowBackgroundData(W^.Data^.At(I));
    GetImage(W^.X1,W^.Y1 + C,W^.X2,W^.Y1 + C + WD^.Lines,WD^.Data^);
    Inc(C,WD^.Lines);
    Inc(I);
  End; { While }
End; { TVGAApplication.SaveBackground }

Procedure TVGAApplication.RestoreBackground(W: PWindowBackground);
Var
  W1,C : LongInt;
  I    : Integer;
  WD   : PWindowBackGroundData;

Begin
  W1 := W^.X2 - W^.X1 + 1;
  C := 0;
  I := 0;
  While I < W^.Data^.Count Do
  Begin
    WD := PWindowBackgroundData(W^.Data^.At(I));
    PutImage(W^.X1,W^.Y1 + C,W1,WD^.Lines,WD^.Data^);
    Inc(C,WD^.Lines);
    Inc(I);
  End; { While }
End; { TVGAApplication.RestoreBackground }

Procedure TVGAApplication.AddForm(Form: PVGAForm);
Begin
  Forms^.Insert(Form);
End; { TVGAApplication.AddForm }

Procedure TVGAApplication.ShowForm(Form: PVGAForm);
Begin
  VGAHideMouse;
  Save(Form^.Left,Form^.Top,Form^.Left + Form^.Width - 1,Form^.Top + Form^.Height - 1);
  Form^.ZOrder := ZOrder + 1;
  If Forms^.Count > 0 Then PVGAForm(Forms^.At(Forms^.Count - 1))^.SetFocused(False);
  SortForms;
  Inc(ZOrder);
  Form^.SetVisible(True);
  Form^.Enabled := True;
  Form^.SetFocused(True);
  Form^.Paint;
  VGAShowMouse;
End; { TVGAApplication.ShowForm }

Procedure TVGAApplication.CloseForm(Form: PVGAForm);
Var
  I        : Integer;
  Form1    : PVGAForm;

Begin
  VGAHideMouse;
  Form^.SetVisible(False);
  Form^.SetFocused(False);
  I := 0;
  While I < Forms^.Count Do
  Begin
    Form1 := PVGAForm(Forms^.At(I));
    If Form1^.ZOrder > Form^.ZOrder Then
    Begin
      Form1^.Paint;
      Dec(Form1^.ZOrder);
    End;
    Inc(I);
  End; { While }
  Form^.ZOrder := -1;
  If Form1^.ZOrder >= 0 Then Form1^.SetFocused(True);
  Dec(ZOrder);
  If Form^.DisposeOnClose Then Forms^.Free(Form);
  SortForms;
  Restore;
  VGAShowMouse;
End; { TVGAApplication.CloseForm }

Procedure TVGAApplication.SortForms;
Var
  I,J         : Integer;
  Form1,Form2 : PVGAForm;

Begin
  I := 0;
  While (I < Forms^.Count - 1) Do
  Begin
    Form1 := PVGAForm(Forms^.At(I));
    J := I + 1;
    While (J < Forms^.Count) Do
    Begin
      Form2 := PVGAForm(Forms^.At(J));
      If Form2^.ZOrder < Form1^.ZOrder Then
      Begin
        Forms^.AtPut(I,Form2);
        Forms^.AtPut(J,Form1);
        Form1 := Form2;
      End;
      Inc(J);
    End; { While }
    Inc(I);
  End; { While }
End; { TVGAApplication.SortForms }

Procedure TVGAApplication.ProcessMessages;
Var Event: TEvent;
Begin
  GetMouseEvent(Event);
  If Event.What = evNothing Then
  Begin
    GetKeyEvent(Event);
    If Event.What = evNothing
     Then Idle
     Else HandleEvent(Event);
  End
  Else HandleEvent(Event);
End; { TVGAApplication.ProcessMessages }

Procedure TVGAApplication.Run;
Begin
  VGAInitEvents;
  Running := True;
  If Forms^.Count > 0 Then
  Begin
    If Assigned(FOnRun) Then FOnRun(@Self,DummyEvent);
    Repeat
      ProcessMessages;
    Until ZOrder < 0;
  End;
  VGADoneEvents;
  Running := False;
End; { TVGAApplication.Run }

Procedure TVGAApplication.Idle;
Begin
End; { TVGAApplication.Idle }

Procedure TVGAApplication.HandleEvent(Var Event: TEvent);
Begin
  Case Event.What Of
    evMouseUp,evMouseDown,evMouseMove,evMouseAuto:
    Begin
      PVGAForm(Forms^.At(Forms^.Count - 1))^.HandleEvent(Event);
    End;
    evKeyDown:
    Begin
      PVGAForm(Forms^.At(Forms^.Count - 1))^.HandleEvent(Event);
    End;
  End; { Case }
End; { TVGAApplication.HandleEvent }

Function TVGAApplication.ShowMessage(Const ACaption,AMessage: String; Buttons: TMsgBoxButton): TMsgBoxResult;
Var MsgBox: PVGAMessageBox;
Begin
  MsgBox := New(PVGAMessageBox,Init(True,ACaption,AMessage,Buttons));
  AddForm(MsgBox);
  If Not Running Then VGAInitEvents;
  ShowForm(MsgBox);
  Repeat
    ProcessMessages;
  Until Not MsgBox^.Visible;
  ShowMessage := MsgBox^.Result;
  Forms^.Free(MsgBox);
  SortForms;
  If Not Running Then VGADoneEvents;
End; { TVGAApplication.ShowMessage }

Function TVGAApplication.InputBox(Const ACaption,AMessage,AInitialText: String): String;
Var IBox: PVGAInputBox;
Begin
  IBox := New(PVGAInputBox,Init(True,ACaption,AMessage,AInitialText));
  AddForm(IBox);
  If Not Running Then VGAInitEvents;
  ShowForm(IBox);
  Repeat
    ProcessMessages;
  Until Not IBox^.Visible;
  InputBox := IBox^.GetText;
  Forms^.Free(IBox);
  SortForms;
  If Not Running Then VGADoneEvents;
End; { TVGAApplication.InputBox }

Function TVGAApplication.FileDialog(Const ACaption,AFilter,AInitialDir: String): String;
Var IBox: PVGAChooser;
Begin
  IBox := New(PVGAChooser,Init(True,ACaption,AFilter,AInitialDir));
  AddForm(IBox);
  If Not Running Then VGAInitEvents;
  ShowForm(IBox);
  Repeat
    ProcessMessages;
  Until Not IBox^.Visible;
  If IBox^.Chose
   Then FileDialog := IBox^.FileName
   Else FileDialog := '';
  Forms^.Free(IBox);
  SortForms;
  If Not Running Then VGADoneEvents;
End; { TVGAApplication.InputBox }

Procedure TVGAApplication.PaintForms;
Var
  I : Integer;
  F : PVGAForm;

Begin
  I := 0;
  While I < Forms^.Count Do
  Begin
    F := PVGAForm(Forms^.At(I));
    If F^.ZOrder >= 0 Then F^.Paint;
    Inc(I);
  End; { While }
End; { TVGAApplication.PaintForms }

Procedure TVGAApplication.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Case E Of
    etOnRun: FOnRun := EventProc;
  End; { Case }
End; { TVGAApplication.AssignEvent }

Procedure TVGAApplication.VGAInitEvents;
Begin
  If Not MouseWasOn Then MouseInit(ScreenWidth,ScreenHeight,2,1,ModeX);
  InitEvents;
  If (ScreenWidth <> 320) Or ModeX Then
  Begin
    HideMouse;
    SetMouseCursor(ClearMask);
  End
  Else SetMouseCursor(ArrowMask);
End; { TVGAApplication.VGAInitEvents }

Procedure TVGAApplication.VGADoneEvents;
Begin
  If (ScreenWidth <> 320) Or ModeX Then
  Begin
    _MouseOff;
    ShowMouse;
  End;
  If Not MouseWasOn Then MouseClose;
  DoneEvents;
  SetMouseCursor(ArrowMask);
End; { TVGAApplication.VGADoneEvents }

{ ----------------------------------------- }
{ TVGAControl                               }
{ ----------------------------------------- }

Constructor TVGAControl.Init(ALeft,ATop,AWidth,AHeight: Integer;
                             AVisible,AEnabled: Boolean; AColor: TVGAColorName);
Begin
  Left     := ALeft;
  Top      := ATop;
  Width    := AWidth;
  Height   := AHeight;
  Visible  := AVisible;
  Enabled  := AEnabled;
  Color    := AColor;
  Parent   := Nil;
  Owner    := Nil;
  Focused  := False;
  TabOrder := -1;
  FOnPaint := Nil;
End; { TVGAControl.Init }

Destructor TVGAControl.Done;
Begin
  Inherited Done;
End; { TVGAControl.Done }

Function TVGAControl.GetAbsLeft: Integer;
Begin
  If Parent <> Nil
   Then GetAbsLeft := Left + Parent^.GetAbsLeft
   Else GetAbsLeft := Left;
End; { TVGAControl.GetAbsLeft }

Function TVGAControl.GetAbsTop: Integer;
Begin
  If Parent <> Nil
   Then GetAbsTop := Top + Parent^.GetAbsTop
   Else GetAbsTop := Top;
End; { TVGAControl.GetABSTop }

Procedure TVGAControl.Paint;
Begin
  Abstract;
End; { TVGAControl.Paint }

Function TVGAControl.GetVisible: Boolean;
Begin
  If Parent <> Nil
   Then GetVisible := Visible And Parent^.GetVisible
   Else GetVisible := Visible;
End; { TVGAControl.GetVisible }

Procedure TVGAControl.SetVisible(V: Boolean);
Begin
  If Parent <> Nil Then Parent^.SetVisibleControl(@Self,V);
  Visible := V;
End; { TVGAControl.SetVisible }

Function TVGAControl.GetFocused: Boolean;
Begin
  GetFocused := Focused;
End; { TVGAControl.GetFocused }

Procedure TVGAControl.SetFocused(V: Boolean);
Begin
  If Parent <> Nil Then Parent^.SetFocusedControl(@Self,V);
  Focused := V;
  Paint;
End; { TVGAControl.SetFocused }

Procedure TVGAControl.SetVisibleControl(C: PVGAControl; V: Boolean);
Begin
End; { TVGAControl.SetVisibleControl }

Procedure TVGAControl.SetFocusedControl(C: PVGAControl; V: Boolean);
Begin
End; { TVGAControl.SetFocusedControl }

Function TVGAControl.GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl;
Begin
  If Any Or (TabOrder >= 0) Then GetControlAt := @Self Else GetControlAt := Nil;
End; { TVGAControl.GetControlAt }

Procedure TVGAControl.HandleEvent(Var Event: TEvent);
Begin
End; { TVGAControl.HandleEvent }

Procedure TVGAControl.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Case E Of
    etOnPaint: FOnPaint := EventProc;
  End; { Case }
End; { TVGAControl.AssignEvent }

Procedure TVGAControl.AddTo(AParent,AOwner: PVGAControl);
Begin
  Parent := AParent;
  Owner  := AOwner;
End; { TVGAControl.AddTo }

Procedure TVGAControl.MouseOff(Var L,T: Integer);
Begin
  L := GetAbsLeft;
  T := GetAbsTop;
  If (MouseWhere.X > L - 16)    Or (MouseWhere.Y > T - 16)      Or
     (MouseWhere.X < L + Width) Or (MouseWhere.Y < T  + Height) Then VGAHideMouse;
End; { TVGAControl.MouseOff }

Function TVGAControl.InBounds(X,Y: Integer): Boolean;
Begin
  InBounds := (X >= Left) And (X < Left + Width) And
              (Y >= Top)  And (Y < Top + Height);
End; { TVGAControl.InBounds }

{ ----------------------------------------- }
{ TVGALabel                                 }
{ ----------------------------------------- }

Constructor TVGALabel.Init(ALeft,ATop,AWidth,AHeight: Integer;
                           AVisible,AEnabled: Boolean; AColor: TVGAColorName; ACaption: String);
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,AVisible,AEnabled,AColor);
  Text := ACaption;
End; { TVGALabel.Init }

Destructor TVGALabel.Done;
Begin
  Inherited Done;
End; { TVGALabel.Done }

Function TVGALabel.GetText: String;
Begin
  GetText := Text;
End; { TVGALabel.GetText }

Procedure TVGALabel.SetText(St: String);
Begin
  Text := St;
  Paint;
End; { TVGALabel.SetText }

Procedure TVGALabel.Paint;
Var
  L,T : Integer;
  C1  : TVGAColorName;

Begin
  If GetVisible Then
  Begin
    MouseOff(L,T);
    If Parent <> Nil Then C1 := Parent^.Color Else C1 := clBtnFace;
    WriteSmOut(L,T,Copy(Text,1,Width Div 6),VGAColors[Color] + VGAColors[C1] Shl 8);
    If Assigned(FOnPaint) Then FOnPaint(@Self,DummyEvent);
    VGAShowMouse;
  End;
End; { TVGALabel.Paint }

{ ----------------------------------------- }
{ TVGAComponent                             }
{ ----------------------------------------- }

Constructor TVGAComponent.Init(ALeft,ATop,AWidth,AHeight: Integer;
                               AVisible,AEnabled: Boolean; AColor: TVGAColorName);
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,AVisible,AEnabled,AColor);
  TabOrder     := 0;
  FOnMouseDown := Nil;
  FOnMouseUp   := Nil;
  FOnMouseMove := Nil;
  FOnKeyDown   := Nil;
End; { TVGAComponent.Init }

Destructor TVGAComponent.Done;
Begin
  Inherited Done;
End; { TVGAComopnent.Done }

Procedure TVGAComponent.HandleEvent(Var Event: TEvent);
Begin
  Case Event.What Of
      evMouseUp: If Assigned(FOnMouseUp)   Then FOnMouseUp(@Self,Event);
    evMouseDown: If Assigned(FOnMouseDown) Then FOnMouseDown(@Self,Event);
    evMouseMove: If Assigned(FOnMouseMove) Then FOnMouseMove(@Self,Event);
      evKeyDown: If Assigned(FOnKeyDown)   Then FOnKeyDown(@Self,Event);
  End; { Case }
End; { TVGAComponent.HandleEvent }

Procedure TVGAComponent.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Inherited AssignEvent(E,EventProc);
  Case E Of
    etOnMouseUp: FOnMouseUp   := EventProc;
  etOnMouseDown: FOnMouseDown := EventProc;
  etOnMouseMove: FOnMouseMove := EventProc;
    etOnKeyDown: FOnKeyDown   := EventProc;
  End; { Case }
End; { TVGAComponent.AssignEvent }

{ ----------------------------------------- }
{ TVGAEdit                                  }
{ ----------------------------------------- }

Constructor TVGAEdit.Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                          AColor: TVGAColorName; Const AText: String);
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,AVisible,AEnabled,AColor);
  Text    := AText;
  TextTop := 1;
  Cursor  := 1;
  Select  := Cursor;
End; { TVGAEdit.Init }

Destructor TVGAEdit.Done;
Begin
  Inherited Done;
End; { TVGAEdit.Done }

Procedure TVGAEdit.Paint;
Var
  L,T : Integer;
  C1  : TVGAColorName;
  C2  : TVGAColorName;
  Len : Integer;
  X   : Integer;
  I,J : Integer;
  P   : Integer;

  Procedure WriteText(Const St: String; Position: Integer; Color: Word);
  Begin
    WriteSmOut(L + 3 + Position * 6,T + (Height Div 2) - 3,St,Color);
  End; { Write }

Begin
  If GetVisible Then
  Begin
    MouseOff(L,T);
    Bevel(L,T,Width,Height,clBtnShadow,clBtnHighlight,cl3DDkShadow,clBtnFace);
    FBox(L + 2,T + 2,L + Width - 3,T + Height - 3,VGAColors[Color]);
    If Select <> Cursor Then
    Begin
      If Select < Cursor Then
      Begin
        I := Select;
        J := Cursor;
      End
      Else
      Begin
        I := Cursor;
        J := Select;
      End;
      If Enabled
       Then C1 := clBtnText
       Else C1 := clGrayText;
      Len := (Width - 6) Div 6;
      P   := TextTop;
      If Enabled
       Then C1 := clBtnText
       Else C1 := clGrayText;
      If I < P Then I := P;
      If J < P Then J := P;
      If I > P + Len Then I := P + Len;
      If J > P + Len Then J := P + Len;
      WriteText(Copy(Text,P,I - P),0,VGAColors[C1] + VGAColors[Color] Shl 8);
      Inc(P,I - P);
      If Enabled Then
      Begin
        C1 := clHighlightText;
        C2 := clHighlight;
      End
      Else
      Begin
        C1 := clGrayText;
        C2 := Color;
      End;
      WriteText(Copy(Text,P,J - I),P - 1,VGAColors[C1] + VGAColors[C2] Shl 8);
      Inc(P,J - I);
      If Enabled
       Then C1 := clBtnText
       Else C1 := clGrayText;
      WriteText(Copy(Text,P,Len - P),P - 1,VGAColors[C1] + VGAColors[Color] Shl 8);
    End
    Else
    Begin
      If Enabled
       Then C1 := clBtnText
       Else C1 := clGrayText;
      Len := (Width - 6) Div 6;
      WriteText(Copy(Text,TextTop,Len),0,VGAColors[C1] + VGAColors[Color] Shl 8);
    End;

    { Draw the cursor }

    If Focused Then
    Begin
      X := L + 3 + (Cursor - TextTop) * 6;
      FBox(X,T + 3,X,T + Height - 4,VGAColors[cl3DDkShadow]);
    End;
    If Assigned(FOnPaint) Then FOnPaint(@Self,DummyEvent);
    VGAShowMouse;
  End;
End; { TVGAEdit.Paint }

Procedure TVGAEdit.HandleEvent(Var Event: TEvent);
Var
  Shift : Byte;
  I,J   : Integer;

Begin
  Case Event.What Of
    evKeyDown:
    Begin
      Shift := GetShiftState;
      Case Event.ScanCode Of
   kP7: Begin                                         { Home }
          If Cursor > 1 Then
          Begin
            If ((Shift And (kbRightShift + kbLeftShift)) <> 0) And
               (Select = Cursor)                               Then Select := Cursor;
            Cursor  := 1;
            TextTop := 1;
            If (Shift And (kbRightShift + kbLeftShift)) = 0 Then Select := Cursor;
            Paint;
          End;
          Event.What := evNothing;
        End;
   kP1: Begin                                         { End }
          If Cursor < Length(Text) + 1 Then
          Begin
            If ((Shift And (kbRightShift + kbLeftShift)) <> 0) And
               (Select = Cursor)                               Then Select := Cursor;
            Cursor := Length(Text) + 1;
            If Cursor > TextTop + ((Width - 6) Div 6) Then
             TextTop := Cursor - ((Width - 6) Div 6);
            If (Shift And (kbRightShift + kbLeftShift)) = 0 Then Select := Cursor;
            Paint;
          End;
          Event.What := evNothing;
        End;
   kP4: Begin                                         { Cursor left }
          If Cursor > 1 Then
          Begin
            If ((Shift And (kbRightShift + kbLeftShift)) <> 0) And
               (Select = Cursor)                               Then Select := Cursor;
            Dec(Cursor);
            If Cursor < TextTop Then TextTop := Cursor;
            If (Shift And (kbRightShift + kbLeftShift)) = 0 Then Select := Cursor;
            Paint;
          End;
          Event.What := evNothing;
        End;
   kP6: Begin                                         { Cursor right }
          If Cursor < Length(Text) + 1 Then
          Begin
            If ((Shift And (kbRightShift + kbLeftShift)) <> 0) And
               (Select = Cursor)                               Then Select := Cursor;
            Inc(Cursor);
            If Cursor > TextTop + ((Width - 6) Div 6) Then
             TextTop := Cursor - ((Width - 6) Div 6);
            If (Shift And (kbRightShift + kbLeftShift)) = 0 Then Select := Cursor;
            Paint;
          End;
          Event.What := evNothing;
        End;
  kDEL: Begin
          If Select <> Cursor Then
          Begin
            If Select < Cursor Then
            Begin
              I := Select;
              J := Cursor;
            End
            Else
            Begin
              I := Cursor;
              J := Select;
            End;
            Delete(Text,I,J - I);
            Select := I;
            Cursor := I;
            Paint;
          End
          Else
          Begin
            If Cursor < Length(Text) + 1 Then
            Begin
              Delete(Text,Cursor,1);
              Paint;
            End;
          End;
          Event.What := evNothing;
        End;
        kBACKSPC:
        Begin
          If Select = Cursor Then
          Begin
            If Cursor > 1 Then
            Begin
              Delete(Text,Cursor - 1,1);
              Dec(Select);
              Dec(Cursor);
              If Cursor < TextTop Then TextTop := Cursor;
            End;
          End
          Else
          Begin
            If Select < Cursor Then
            Begin
              Delete(Text,Select,Cursor - Select);
              Cursor := Select;
              If Cursor < TextTop Then TextTop := Cursor;
            End
            Else
            Begin
              Delete(Text,Cursor,Select - Cursor);
              Select := Cursor;
            End;
          End;
          Paint;
        End;
      Else
        Case Event.CharCode Of
          #32..#255:
          Begin
            If Select <> Cursor Then
            Begin
              If Select < Cursor Then
              Begin
                I := Select;
                J := Cursor;
              End
              Else
              Begin
                I := Cursor;
                J := Select;
              End;
              Delete(Text,I,J - I);
              Select := I;
              Cursor := I;
            End;
            Insert(Event.CharCode,Text,Cursor);
            Inc(Cursor);
            Inc(Select);
            If Cursor > TextTop + ((Width - 6) Div 6) Then Inc(TextTop);
            Paint;
            Event.What := evNothing;
          End;
        End; { Case }
      End; { Case }
    End;
  End; { Case }
  Inherited HandleEvent(Event);
End; { TVGAEdit.HandleEvent }

Procedure TVGAEdit.SetText(St: String);
Begin
  Text    := St;
  Cursor  := 1;
  TextTop := 1;
  Select  := Cursor;
  Paint;
End; { TVGAEdit.SetText }

Function TVGAEdit.GetText: String;
Begin
  GetText := Text;
End; { TVGAEdit.GetText }

{ ----------------------------------------- }
{ TVGAButton                                }
{ ----------------------------------------- }

Constructor TVGAButton.Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                            AColor: TVGAColorName; Const ACaption: String);
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,AVisible,AEnabled,AColor);
  Caption  := ACaption;
  Pressed  := False;
  Color    := AColor;
  FOnClick := Nil;
End; { TVGAButton.Init }

Destructor TVGAButton.Done;
Begin
  Inherited Done;
End; { TVGAButton.Done }

Procedure TVGAButton.Paint;
Var
  L,T : Integer;
  C1  : TVGAColorName;
  C2  : TVGAColorName;
  I,J : Integer;
  B   : Boolean;
  X,Y : Integer;

Begin
  If GetVisible Then
  Begin
    MouseOff(L,T);

    FBox(L + 1,T + 1,L + Width - 2,T + Height - 2,VGAColors[Color]);

    If Focused Then
    Begin
      If Pressed Then J := 1 Else J := 0;
      B := False;
      X := Width - 7;
      Y := Height - 7;
      For I := 0 To (X + Y) * 2 - 1 Do
      Begin
        If B Then
        Begin
          If I < X Then PutPixel(L + 3 + I + J,T + 3 + J,VGAColors[cl3DDkShadow]) Else
           If I < X + Y Then PutPixel(L + Width - 4 + J,T + 3 + I - X + J,VGAColors[cl3DDkShadow]) Else
            If I < 2 * X + Y Then PutPixel(L + Width - 4 - (I - (X + Y)) + J,T + Height - 4 + J,VGAColors[cl3DDkShadow]) Else
             PutPixel(L + 3 + J,T + Height - 4 - (I - (2 * X + Y)) + J,VGAColors[cl3DDkShadow]);
        End;
        B := Not B;
      End; { For I }
    End;
    If Pressed
     Then Bevel(L,T,Width,Height,cl3dDkShadow,clBtnHighlight,clBtnShadow,Color)
     Else Bevel(L,T,Width,Height,clBtnHighlight,cl3dDkShadow,Color,clBtnShadow);
    If Enabled
     Then C1 := clBtnText
     Else C1 := clGrayText;
    If Pressed Then I := 1 Else I := 0;
    If Caption <> '' Then
     WriteSmOut(L + (Width Div 2) - 3 * Length(Caption) + I,T + (Height Div 2) - 3 + I,
                Caption,VGAColors[C1] + VGAColors[Color] Shl 8);
    If Assigned(FOnPaint) Then FOnPaint(@Self,DummyEvent);
    VGAShowMouse;
  End;
End; { TVGAButton.Paint }

Procedure TVGAButton.HandleEvent(Var Event: TEvent);
Begin
  Case Event.What Of
    evMouseUp:
    Begin
      Pressed := False;
      Paint;
      If Assigned(FOnClick) Then FOnClick(@Self,Event);
    End;
    evMouseDown:
    Begin
      Pressed := True;
      Paint;
    End;
    evKeyDown:
    Begin
      If Event.CharCode In [#13,#32] Then
      Begin
        Pressed := True;
        Paint;
        Pressed := False;
        Paint;
        If Assigned(FOnClick) Then FOnClick(@Self,Event);
        Event.What := evNothing;  { Dispatch event }
      End;
    End;
  End; { Case }
  Inherited HandleEvent(Event);
End; { TVGAButton.HandleEvent }

Procedure TVGAButton.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Inherited AssignEvent(E,EventProc);
  Case E Of
    etOnClick: FOnClick := EventProc;
  End; { Case }
End; { TVGAButton.AssignEvent }

{ ----------------------------------------- }
{ TVGASpeedButton                           }
{ ----------------------------------------- }

Constructor TVGASpeedButton.Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                                 AColor: TVGAColorName);
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,AVisible,AEnabled,AColor,'');
End; { TVGASpeedButton.Init }

Destructor TVGASpeedButton.Done;
Begin
  Inherited Done;
End; { TVGASpeedButton.Done }

Procedure TVGASpeedButton.Paint;
Var
  L,T : Integer;
  C1  : TVGAColorName;
  C2  : TVGAColorName;
  I   : Integer;

Begin
  If GetVisible Then
  Begin
    MouseOff(L,T);
    FBox(L + 1,T + 1,L + Width - 2,T + Height - 2,VGAColors[Color]);
    If Pressed
     Then Bevel(L,T,Width,Height,clBtnShadow,clBtnHighlight,Color,Color)
     Else Bevel(L,T,Width,Height,clBtnShadow,cl3dDkShadow,clBtnHighlight,clBtnShadow);
    If Assigned(FOnPaint) Then FOnPaint(@Self,DummyEvent);
    VGAShowMouse;
  End;
End; { TVGASpeedButton.Paint }

Procedure TVGASpeedButton.HandleEvent(Var Event: TEvent);
Begin
  Inherited HandleEvent(Event);
End; { TVGASpeedButton.HandleEvent }

Procedure TVGASpeedButton.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Inherited AssignEvent(E,EventProc);
End; { TVGASpeedButton.AssignEvent }

{ ----------------------------------------- }
{ TVGAScrollBar                             }
{ ----------------------------------------- }

Procedure TVGAScrollBar_btnUpClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAScrollBar(Sender^.Parent)^.btnUpClick(Sender^.Parent,Event);
End; { TVGAScrollBar_btnUpClick }

Procedure TVGAScrollBar_btnDownClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAScrollBar(Sender^.Parent)^.btnDownClick(Sender^.Parent,Event);
End; { TVGAScrollBar_btnDownClick }

Procedure TVGAScrollBar_btnUpPaint(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAScrollBar(Sender^.Parent)^.btnUpPaint(Sender^.Parent,Event);
End; { TVGAScrollBar_btnUpPaint }

Procedure TVGAScrollBar_btnDownPaint(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAScrollBar(Sender^.Parent)^.btnDownPaint(Sender^.Parent,Event);
End; { TVGAScrollBar_btnDownPaint }

Procedure TVGAScrollBar.btnUpClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  If Position > Min Then
  Begin
    Dec(Position);
    Paint;
    If Assigned(FOnChange) Then FOnChange(Sender,Event);
  End;
End; { TVGAScrollBar.btnUpClick }

Procedure TVGAScrollBar.btnDownClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  If Position < Max Then
  Begin
    Inc(Position);
    Paint;
    If Assigned(FOnChange) Then FOnChange(Sender,Event);
  End;
End; { TVGAScrollBar.btnDownClick }

Procedure TVGAScrollBar.btnUpPaint(Sender: PVGAControl; Var Event: TEvent);
Var
  L,T : Integer;
  X,Y : Integer;
  C   : TVGAColorName;

Begin
  MouseOff(L,T);
  If Enabled
   Then C := clBtnText
   Else C := clGrayText;
  If Orientation = orHorizontal Then
  Begin
    X := L + Height Div 2;
    Y := T + Height Div 2;
    PutPixel(X - 1,Y,VGAColors[C]);
    FBox(X,Y - 1,X,Y + 1,VGAColors[C]);
    FBox(X + 1,Y - 2,X + 1,Y + 2,VGAColors[C]);
  End
  Else
  Begin
    X := L + Width Div 2;
    Y := T + Width Div 2;
    PutPixel(X,Y - 1,VGAColors[C]);
    FBox(X - 1,Y,X + 1,Y,VGAColors[C]);
    FBox(X - 2,Y + 1,X + 2,Y + 1,VGAColors[C]);
  End;
  VGAShowMouse;
End; { TVGAScrollBar.btnUpPaint }

Procedure TVGAScrollBar.btnDownPaint(Sender: PVGAControl; Var Event: TEvent);
Var
  L,T : Integer;
  X,Y : Integer;
  C   : TVGAColorName;

Begin
  MouseOff(L,T);
  If Enabled
   Then C := clBtnText
   Else C := clGrayText;
  If Orientation = orHorizontal Then
  Begin
    X := L + Width - (Height Div 2) - 1;
    Y := T + Height Div 2;
    PutPixel(X + 1,Y,VGAColors[C]);
    FBox(X,Y - 1,X,Y + 1,VGAColors[C]);
    FBox(X - 1,Y - 2,X - 1,Y + 2,VGAColors[C]);
  End
  Else
  Begin
    X := L + Width Div 2;
    Y := T + Height - (Width Div 2) - 1;
    PutPixel(X,Y + 1,VGAColors[C]);
    FBox(X - 1,Y,X + 1,Y,VGAColors[C]);
    FBox(X - 2,Y - 1,X + 2,Y - 1,VGAColors[C]);
  End;
  VGAShowMouse;
End; { TVGAScrollBar.btnDownPaint }

Constructor TVGAScrollBar.Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                               AColor: TVGAColorName; AType: TOrientation; AMin,AMax,APosition,ALargeChange: LongInt);
Var W,L,T: Integer;
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,AVisible,AEnabled,AColor);
  FOnChange   := Nil;
  Min         := AMin;
  Max         := AMax;
  Position    := APosition;
  LargeChange := ALargeChange;
  Orientation := AType;
  If AType = orHorizontal Then
  Begin
    W := AHeight;
    L := AWidth - W;
    T := 0;
  End
  Else
  Begin
    W := AWidth;
    L := 0;
    T := AHeight - W;
  End;
  btnUp   := New(PVGASpeedButton,Init(0,0,W,W,True,True,clBtnFace));
  btnDown := New(PVGASpeedButton,Init(L,T,W,W,True,True,clBtnFace));
  btnUp^.AssignEvent(etOnClick,TVGAScrollBar_btnUpClick);
  btnDown^.AssignEvent(etOnClick,TVGAScrollBar_btnDownClick);
  btnUp^.AssignEvent(etOnPaint,TVGAScrollBar_btnUpPaint);
  btnDown^.AssignEvent(etOnPaint,TVGAScrollBar_btnDownPaint);
End; { TVGAScrollBar.Init }

Destructor TVGAScrollBar.Done;
Begin
  Dispose(btnUp,Done);
  Dispose(btnDown,Done);
  Inherited Done;
End; { TVGAScrollBar.Done }

Procedure TVGAScrollBar.Paint;
Var
  L,T       : Integer;
  L1,T1,W,H : Integer;
  W1,H1     : Integer;
  I         : Integer;

Begin
  If GetVisible Then
  Begin
    MouseOff(L,T);
    If Orientation = orHorizontal Then
    Begin
      L1 := L + Height;
      T1 := T;
      W  := Width - 2 * Height;
      H  := Height;
      W1 := W;
      H1 := H;
    End
    Else
    Begin
      L1 := L;
      T1 := T + Width;
      W  := Width;
      H  := Height - 2 * Width;
      W1 := H;
      H1 := W;
    End;
    FBox(L1,T1,L1 + W - 1,T1 + H - 1,VGAColors[Color]);
    btnUp^.Paint;
    btnDown^.Paint;
    If Max > Min
     Then I := Trunc((Position  - Min) / (Max - Min) * W1 - (H1 Div 2))
     Else I := 0;
    If I < 0 Then I := 0;
    If I > W1 - H1 Then I := W1 - H1;
    If Orientation = orHorizontal Then
    Begin
      L1 := L + H1 + I;
      T1 := T;
    End
    Else
    Begin
      L1 := L;
      T1 := T + H1 + I;
    End;
    Bevel(L1,T1,H1,H1,clBtnShadow,cl3DDkShadow,clBtnHighlight,clBtnShadow);
    FBox(L1 + 2,T1 + 2,L1 + H1 - 3,T1 + H1 - 3,VGAColors[clBtnFace]);
    If Assigned(FOnPaint) Then FOnPaint(@Self,DummyEvent);
    VGAShowMouse;
  End;
End; { TVGAScrollBar.Paint }

Procedure TVGAScrollBar.HandleEvent(Var Event: TEvent);
Var
  L,T       : Integer;
  L1,T1,W,H : Integer;
  W1,H1     : Integer;
  I,K       : Integer;
  X,Y       : Integer;

Begin
  Case Event.What Of
    evMouseUp:
    Begin
      L := GetAbsLeft;
      T := GetAbsTop;
      If Orientation = orHorizontal Then
      Begin
        L1 := L + Height;
        T1 := T;
        W  := Width - 2 * Height;
        H  := Height;
        W1 := W;
        H1 := H;
      End
      Else
      Begin
        L1 := L;
        T1 := T + Width;
        W  := Width;
        H  := Height - 2 * Width;
        W1 := H;
        H1 := W;
      End;
      If Max > Min
       Then I := Trunc((Position  - Min) / (Max - Min) * W1 - (H1 Div 2))
       Else I := 0;
      If I < 0 Then I := 0;
      If I > W1 - H1 Then I := W1 - H1;
      If Orientation = orHorizontal Then
      Begin
        X := Event.Where.X - L;
        Y := Event.Where.Y - T;
      End
      Else
      Begin
        Y := Event.Where.X - L;
        X := Event.Where.Y - T;
      End;
      If (Y >= 0) And (Y < H1) Then
      Begin
        If (X > H1) And (X < I + H1) Then
        Begin
          Dec(Position,LargeChange);
          If Position < Min Then Position := Min;
        End
        Else
         If (X > I + 2 * H1) And (X < W1 + H1) Then
         Begin
           Inc(Position,LargeChange);
           If Position > Max Then Position := Max;
         End;
      End;
      Paint;
      If Assigned(FOnChange) Then FOnChange(@Self,Event);
    End;
  End; { Case }
  Inherited HandleEvent(Event);
End; { TVGAScrollBar.HandleEvent }

Procedure TVGAScrollBar.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Inherited AssignEvent(E,EventProc);
  Case E Of
    etOnChange: FOnChange := EventProc;
  End; { Case }
End; { TVGAScrollBar.AssignEvent }

Procedure TVGAScrollBar.AddTo(AParent,AOwner: PVGAControl);
Begin
  Inherited AddTo(AParent,AOwner);
  btnUp^.AddTo(@Self,AOwner);
  btnDown^.AddTo(@Self,AOwner);
End; { TVGAScrollBar.AddTo }

Function TVGAScrollBar.GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl;
Var C: PVGAControl;
Begin
  If btnUp^.InBounds(X,Y)
    Then C := btnUp^.GetControlAt(X - btnUp^.Left,Y - btnUp^.Top,Any)
    Else
     If btnDown^.InBounds(X,Y)
      Then C := btnDown^.GetControlAt(X - btnDown^.Left,Y - btnDown^.Top,Any)
      Else C := Inherited GetControlAt(X,Y,False);
  GetControlAt := C;
End; { TVGAScrollBar.GetControlAt }

{ ----------------------------------------- }
{ TVGAListBox                               }
{ ----------------------------------------- }

Procedure TVGAListBox_sbSelectChange(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAListBox(Sender^.Parent)^.sbSelectChange(Sender^.Parent,Event);
End; { TVGAListBox_sbSelectChange }

Procedure TVGAListBox.sbSelectChange(Sender: PVGAControl; Var Event: TEvent);
Var LB: PVGAListBox;
Begin
  Paint;
  If Assigned(FOnChange) Then FOnChange(Sender,Event);
End; { TVGAListBox.sbSelectChange }

Constructor TVGAListBox.Init(ALeft,ATop,AWidth,AHeight: Integer; AVisible,AEnabled: Boolean;
                             AColor: TVGAColorName; AType: TOrientation; AColWidth,ARowHeight: Integer);
Var L,T,W,H : Integer;
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,AVisible,AEnabled,AColor);
  FOnChange   := Nil;
  FOnDblClick := Nil;
  Orientation := AType;
  ColWidth    := AColWidth;
  RowHeight   := ARowHeight;
  ItemIndex   := -1;  { Nothing initially selected }
  If Orientation = orHorizontal Then
  Begin
    L := 2;
    T := AHeight - ScrollBarSize - 2;
    W := AWidth - 4;
    H := ScrollBarSize;
  End
  Else
  Begin
    L := AWidth - ScrollBarSize - 2;
    T := 2;
    W := ScrollBarSize;
    H := AHeight - 4;
  End;
  sbSelect := New(PVGAScrollBar,Init(L,T,W,H,True,True,clScrollBar,Orientation,0,100,0,10));
  sbSelect^.AssignEvent(etOnChange,TVGAListBox_sbSelectChange);
  Items    := New(PCollection,Init(10,10));
End; { TVGAListBox.Init }

Destructor TVGAListBox.Done;
Begin
  Dispose(sbSelect,Done);
  Dispose(Items,Done);
  Inherited Done;
End; { TVGAListBox.Done }

Procedure TVGAListBox.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Inherited AssignEvent(E,EventProc);
  Case E Of
    etOnChange: FOnChange   := EventProc;
  etOnDblClick: FOnDblClick := EventProc;
  End; { Case }
End; { TVGAListBox.AssignEvent }

Procedure TVGAListBox.AddTo(AParent,AOwner: PVGAControl);
Begin
  Inherited AddTo(AParent,AOwner);
  sbSelect^.AddTo(@Self,AOwner);
End; { TVGAListBox.AddTo }

Function TVGAListBox.GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl;
Var C: PVGAControl;
Begin
  If sbSelect^.InBounds(X,Y)
   Then C := sbSelect^.GetControlAt(X - sbSelect^.Left,Y - sbSelect^.Top,Any)
   Else C := Inherited GetControlAt(X,Y,False);
  GetControlAt := C;
End; { TVGAListBox.GetControlAt }

Procedure TVGAListBox.Paint;
Var
  L,T  : Integer;
  Cols : Integer;
  Rows : Integer;
  I,J  : Integer;
  CF   : TVGAColorName;
  CB   : TVGAColorName;
  Len  : Integer;
  X,Y  : Integer;
  X2   : Integer;

Begin
  If GetVisible Then
  Begin
    MouseOff(L,T);
    Bevel(L,T,Width,Height,clBtnShadow,clBtnHighlight,cl3DDkShadow,clBtnShadow);

    If Orientation = orHorizontal Then
    Begin
      Rows := (Height - ScrollBarSize - 4) Div RowHeight;
      Cols := (Width - 4) Div ColWidth;
      If Cols * ColWidth < Width - 4 Then Inc(Cols);
      FBox(L + 2,T + 2,L + Width - 3,T + Height - ScrollBarSize - 3,VGAColors[Color]);
      I := sbSelect^.Position * Rows;
      J := 0;
      While (J < Rows * Cols) And (J + sbSelect^.Position < Items^.Count) Do
      Begin
        If (I = ItemIndex) And Enabled Then
        Begin
          If Focused Then
          Begin
            CF := clHighlightText;
            CB := clHighlight;
          End
          Else
          Begin
            CF := clNoHighlightText;
            CB := clNoHighlight;
          End;
        End
        Else
        Begin
          If Enabled Then CF := clBtnText Else CF := clGrayText;
          CB := Color;
        End;
        X   := (J Div Rows) * ColWidth;
        Y   := (J Mod Rows) * RowHeight;
        Len := Width - 4 - X - 2;
        If Len > ColWidth - 2 Then Len := ColWidth - 2;
        Len := Len Div 6;
        X2  := X + ColWidth - 1;
        If X2 > Width - 5 Then X2 := Width - 5;
        FBox(L + 2 + X,T + 2 + Y,L + 2 + X2,T + 2 + Y + RowHeight - 1,VGAColors[CB]);
        If I < Items^.Count Then
         WriteSmOut(L + 4 + X,T + 4 + Y,Copy(StrPas(PStringObj(Items^.At(I))^.P),1,Len),VGAColors[CF] + VGAColors[CB] Shl 8);
        Inc(J);
        Inc(I);
      End; { While }
    End
    Else
    Begin
      FBox(L + 2,T + 2,L + Width - ScrollBarSize - 3,T + Height - 3,VGAColors[Color]);
      J    := (Width - ScrollBarSize - 4) Div 6;
      I    := sbSelect^.Position;
      Rows := (Height - 4) Div RowHeight;
      I    := 0;
      While (I < Rows) And (I + sbSelect^.Position < Items^.Count) Do
      Begin
        If (I + sbSelect^.Position = ItemIndex) And Enabled Then
        Begin
          CF := clHighlightText;
          CB := clHighlight;
        End
        Else
        Begin
          If Enabled Then CF := clBtnText Else CF := clGrayText;
          CB := Color;
        End;
        X := L + 2;
        Y := T + 2 + I * RowHeight;
        FBox(X,Y,L + Width - ScrollBarSize - 3,Y + RowHeight - 1,VGAColors[CB]);
        WriteSmOut(X,Y + (RowHeight - 8) Div 2,
                   Copy(StrPas(PStringObj(Items^.At(I + sbSelect^.Position))^.P),1,J),VGAColors[CF] + VGAColors[CB] Shl 8);
        Inc(I);
      End; { While }
    End;
    sbSelect^.Paint;
    VGAShowMouse;
  End;
End; { TVGAListBox.Paint }

Procedure TVGAListBox.HandleEvent(Var Event: TEvent);
Var
  L,T       : Integer;
  X,Y       : Integer;
  Rows      : Integer;
  Cols      : Integer;
  R,C       : Integer;
  I         : Integer;
  WholeCols : Integer;
  Count     : Integer;
  Position  : Integer;

Begin
  Count    := Items^.Count;
  Position := sbSelect^.Position;
  If Orientation = orHorizontal Then
  Begin
    Rows := (Height - ScrollBarSize - 4) Div RowHeight;
    Cols := (Width - 4) Div ColWidth;
    WholeCols := Cols;
    If Cols * ColWidth < Width - 4 Then Inc(Cols);
  End
  Else Rows := (Height - 4) Div RowHeight;
  Case Event.What Of
    evMouseUp:
    Begin
      L := GetAbsLeft;
      T := GetAbsTop;
      X := Event.Where.X - L - 2;
      Y := Event.Where.Y - T - 2;
      If (X >= 0) And (Y >= 0) Then
      Begin
        If Orientation = orHorizontal Then
        Begin
          R := Y Div RowHeight;
          C := X Div ColWidth;
          If (R < Rows) And (C < Cols) And (X < Width - 4) Then
          Begin
            I := (C + Position) * Rows + R;
            If I < Count Then
            Begin
              If I <> ItemIndex Then
              Begin
                ItemIndex := I;
                Paint;
                If Assigned(FOnChange) Then FOnChange(@Self,Event);
              End;
            End;
          End;
        End
        Else
        Begin
          R := Y Div RowHeight;
          If R < Rows Then
          Begin
            I := R + Position;
            If (I <> ItemIndex) And (I < Count) Then
            Begin
              ItemIndex := I;
              Paint;
              If Assigned(FOnChange) Then FOnChange(@Self,Event);
            End;
          End;
        End;
      End;
      If Event.Double Then
      Begin
        If Assigned(FOnDblClick) Then FOnDblClick(@Self,Event);
      End;
    End;
    evkeyDown:
    Begin
      Case Event.CharCode Of
        #0:
        Begin
          Case Event.ScanCode Of
       kP7: Begin                                         { Home }
              If ItemIndex <> 0 Then
              Begin
                ItemIndex          := 0;
                sbSelect^.Position := 0;
                Paint;
              End;
            End;
       kP1: Begin                                         { End }
              If ItemIndex < Count Then
              Begin
                If ItemIndex < Count - 1 Then
                Begin
                  ItemIndex := Count - 1;
                  If Orientation = orHorizontal Then
                  Begin
                    While (ItemIndex Div Rows) >= sbSelect^.Position + WholeCols Do Inc(sbSelect^.Position);
                  End
                  Else
                  Begin
                    While ItemIndex >= sbSelect^.Position + Rows Do Inc(sbSelect^.Position);
                  End;
                  If Assigned(FOnChange) Then FOnChange(@Self,Event);
                  Paint;
                End;
              End;
            End;
       kP4: Begin                                         { Cursor left }
              If (ItemIndex < 0) And (Count > 0) Then
              Begin
                ItemIndex          := 0;
                sbSelect^.Position := 0;
                If Assigned(FOnChange) Then FOnChange(@Self,Event);
                Paint;
              End
              Else
              Begin
                If (ItemIndex > 0) And (Orientation = orHorizontal) Then
                Begin
                  Dec(ItemIndex,Rows);
                  If ItemIndex < 0 Then ItemIndex := 0;
                  If (ItemIndex Div Rows) < Position Then Dec(sbSelect^.Position);
                  If Assigned(FOnChange) Then FOnChange(@Self,Event);
                  Paint;
                End;
              End;
            End;
       kP6: Begin                                         { Cursor right }
              If (ItemIndex < 0) And (Count > 0) Then
              Begin
                ItemIndex          := 0;
                sbSelect^.Position := 0;
                If Assigned(FOnChange) Then FOnChange(@Self,Event);
                Paint;
              End
              Else
              Begin
                If (ItemIndex < Count - 1) And (Orientation = orHorizontal) Then
                Begin
                  Inc(ItemIndex,Rows);
                  If ItemIndex >= Count Then ItemIndex := Count - 1;
                  If (ItemIndex Div Rows) >= Position + WholeCols Then Inc(sbSelect^.Position);
                  If Assigned(FOnChange) Then FOnChange(@Self,Event);
                  Paint;
                End;
              End;
            End;
       kP8: Begin                                         { Cursor up }
              If (ItemIndex < 0) And (Count > 0) Then
              Begin
                ItemIndex          := 0;
                sbSelect^.Position := 0;
                If Assigned(FOnChange) Then FOnChange(@Self,Event);
                Paint;
              End
              Else
              Begin
                If ItemIndex > 0 Then
                Begin
                  Dec(ItemIndex);
                  If Orientation = orHorizontal Then
                  Begin
                    If (ItemIndex Div Rows) < Position Then Dec(sbSelect^.Position);
                  End
                  Else
                  Begin
                    If ItemIndex < Position Then sbSelect^.Position := ItemIndex;
                  End;
                  If Assigned(FOnChange) Then FOnChange(@Self,Event);
                  Paint;
                End;
              End;
            End;
       kP2: Begin                                         { Cursor down }
              If (ItemIndex < 0) And (Count > 0) Then
              Begin
                ItemIndex          := 0;
                sbSelect^.Position := 0;
                If Assigned(FOnChange) Then FOnChange(@Self,Event);
                Paint;
              End
              Else
              Begin
                If ItemIndex < Count - 1 Then
                Begin
                  Inc(ItemIndex);
                  If Orientation = orHorizontal Then
                  Begin
                    If (ItemIndex Div Rows) >= Position + WholeCols Then Inc(sbSelect^.Position);
                  End
                  Else
                  Begin
                    If ItemIndex >= Position + Rows Then Inc(sbSelect^.Position);
                  End;
                  If Assigned(FOnChange) Then FOnChange(@Self,Event);
                  Paint;
                End;
              End;
            End;
       kP9: Begin                                         { Page up }
              If (ItemIndex < 0) And (Count > 0) Then
              Begin
                ItemIndex          := 0;
                sbSelect^.Position := 0;
                If Assigned(FOnChange) Then FOnChange(@Self,Event);
                Paint;
              End
              Else
              Begin
                If ItemIndex > 0 Then
                Begin
                  If Orientation = orHorizontal Then
                  Begin
                    Dec(ItemIndex,Rows * WholeCols);
                    If ItemIndex < 0 Then ItemIndex := 0;
                    If (ItemIndex Div Rows) < Position Then Dec(sbSelect^.Position,WholeCols);
                  End
                  Else
                  Begin
                    Dec(ItemIndex,Rows);
                    If ItemIndex < 0 Then ItemIndex := 0;
                    If ItemIndex < Position Then sbSelect^.Position := ItemIndex;
                  End;
                  If Assigned(FOnChange) Then FOnChange(@Self,Event);
                  Paint;
                End;
              End;
            End;
       kP3: Begin                                         { Page down }
              If (ItemIndex < 0) And (Count > 0) Then
              Begin
                ItemIndex          := 0;
                sbSelect^.Position := 0;
                If Assigned(FOnChange) Then FOnChange(@Self,Event);
                Paint;
              End
              Else
              Begin
                If ItemIndex < Count - 1 Then
                Begin
                  If Orientation = orHorizontal Then
                  Begin
                    Inc(ItemIndex,Rows * WholeCols);
                    If ItemIndex >= Count Then ItemIndex := Count - 1;
                    If (ItemIndex Div Rows) >= Position + WholeCols Then Inc(sbSelect^.Position,WholeCols);
                  End
                  Else
                  Begin
                    Inc(ItemIndex,Rows);
                    If ItemIndex >= Count Then ItemIndex := Count - 1;
                    If ItemIndex >= Position + Rows Then sbSelect^.Position := ItemIndex - Rows + 1;
                  End;
                  If Assigned(FOnChange) Then FOnChange(@Self,Event);
                  Paint;
                End;
              End;
            End;
          End; { Case }
          Event.What := evNothing;
        End;
      End; { Case }
    End;
  End; { Case }
  Inherited HandleEvent(Event);
End; { TVGAListBox.HandleEvent }

Procedure TVGAListBox.AddString(St: String);
Var
  Rows : Integer;
  Cols : Integer;
  I    : Integer;

Begin
  Items^.Insert(New(PStringObj,Init(St)));
  I := Items^.Count;
  If Orientation = orHorizontal Then
  Begin
    Rows := (Height - ScrollBarSize - 4) Div RowHeight;
    Cols := (Width - 4) Div ColWidth;
    If I <= Rows * Cols
     Then I := 0
     Else I := ((I - Rows * Cols) Div Rows) + 1;
    sbSelect^.Max := I;
  End
  Else
  Begin
    Rows := (Height - 4) Div RowHeight;
    If I <= Rows
     Then I := 0
     Else I := (I - Rows);
    sbSelect^.Max := I;
  End;
  sbSelect^.Min := 0;
  If Orientation = orHorizontal
   Then sbSelect^.LargeChange := (Width - 4) Div ColWidth
   Else sbSelect^.LargeChange := (Height - ScrollBarSize - 4) Div RowHeight;
End; { TVGAListBox.AddString }

Procedure TVGAListBox.Clear;
Begin
  Items^.FreeAll;
  ItemIndex          := -1;
  sbSelect^.Position := 0;
  sbSelect^.Max      := 0;
  Paint;
End; { TVGAListBox.Clear }

{ ----------------------------------------- }
{ TVGAForm                                  }
{ ----------------------------------------- }

Constructor TVGAForm.Init(ALeft,ATop,AWidth,AHeight: Integer; AEnabled: Boolean;
                          AColor: TVGAColorName; Const ACaption: String; ABorder: TBorderStyle);
Begin
  Inherited Init(ALeft,ATop,AWidth,AHeight,False,AEnabled,AColor);
  Caption          := ACaption;
  Controls         := New(PCollection,Init(10,10));
  BorderStyle      := ABorder;
  TabNum           := 0;
  MaxTab           := 0;
  ZOrder           := -1;
  Modal            := False;
  FocusedComponent := -1;
  Updating         := False;
  Default          := Nil;
  FOnShow          := Nil;
  FOnClose         := Nil;
  DisposeOnClose   := False;
End; { TVGAForm.Init }

Destructor TVGAForm.Done;
Begin
  Dispose(Controls,Done);
  Inherited Done;
End; { TVGAForm.Done }

Procedure TVGAForm.AssignEvent(E: TEventType; EventProc: TNotifyEvent);
Begin
  Inherited AssignEvent(E,EventProc);
  Case E Of
    etOnShow: FOnShow  := EventProc;
   etOnClose: FOnClose := EventProc;
  End; { Case }
End; { TVGAForm.AssignEvent }

Function TVGAForm.ClientWidth: Integer;
Begin
  Case BorderStyle Of
    bsNone: ClientWidth := Width;
  bsSingle: ClientWidth := Width - 4;
  bsDialog: ClientWidth := Width - 2 * BorderWidth - 4;
  End; { Case }
End; { TVGAForm.ClientWidth }

Function TVGAForm.ClientHeight: Integer;
Begin
  Case BorderStyle Of
    bsNone: ClientHeight := Height;
  bsSingle: ClientHeight := Height - 4;
  bsDialog: ClientHeight := Height - 2 * BorderWidth - 4;
  End; { Case }
End; { TVGAForm.ClientHeight }

Function TVGAForm.GetAbsLeft: Integer;
Var L: Integer;
Begin
  Case BorderStyle Of
    bsNone: L := 0;
  bsSingle: L := 1;
  bsDialog: L := BorderWidth + 2;
  End; { Case }
  GetAbsLeft := Inherited GetAbsLeft + L;
End; { TVGAForm.GetAbsLeft }

Function TVGAForm.GetAbsTop: Integer;
Var T: Integer;
Begin
  Case BorderStyle Of
    bsNone: T := 0;
  bsSingle: T := 1;
  bsDialog: T := BorderWidth + CaptionHeight + 2;
  End; { Case }
  GetAbsTop := Inherited GetAbsTop + T;
End; { TVGAForm.GetAbsTop }

Procedure TVGAForm.AddControl(Control: PVGAControl);
Begin
  Controls^.Insert(Control);
  Control^.AddTo(@Self,@Self);
  If Control^.TabOrder >= 0 Then
  Begin
    If MaxTab = TabNum Then FocusedComponent := Controls^.Count - 1;
    Control^.TabOrder := MaxTab;
    Control^.SetFocused(MaxTab = TabNum);
    Inc(MaxTab);
  End;
End; { TVGAForm.AddControl }

Procedure TVGAForm.Paint;
Var I,J,CL,CT,CW,L,T: Integer;
Begin
  If GetVisible Then
  Begin
    MouseOff(L,T);

    { Draw outside border, if any }

    If BorderStyle <> bsNone Then
    Begin
      Bevel(Left,Top,Width,Height,clBtnFace,cl3DDkShadow,cl3DLight,clBtnShadow);
    End;

    { Draw dialog border }

    If BorderStyle = bsDialog Then
    Begin
      FBox(Left + 2,Top + 2,Left + BorderWidth + 1,Top + Height - 3,VGAColors[clActiveBorder]);
      FBox(Left + 2,Top + 2,Left + Width - 3,Top + BorderWidth + 1,VGAColors[clActiveBorder]);
      FBox(Left + 2,Top + Height - BorderWidth - 2,Left + Width - 3,Top + Height - 3,VGAColors[clActiveBorder]);
      FBox(Left + Width - BorderWidth - 2,Top + 2,Left + Width - 3,Top + Height - 3,VGAColors[clActiveBorder]);
    End;

    { Fill center of window }

    Case BorderStyle Of
       bsNone: Begin
                 I := 0;
                 J := 0;
               End;
     bsSingle: Begin
                 I := 2;
                 J := 2;
               End;
     bsDialog: Begin
                 I := BorderWidth + 2;
                 J := BorderWidth + CaptionHeight + 2;
               End;
    End; { Case }
    FBox(Left + I,Top + J,Left + Width - I - 1,Top + Height - I - 1,VGAColors[Color]);

    { Draw caption }

    If BorderStyle = bsDialog Then
    Begin
      CL := Left + BorderWidth + 2;
      CT := Top  + BorderWidth + 2;
      CW := Width - 2 * (BorderWidth + 2);
      FBox(CL,CT,CL + CW - 1,CT + CaptionHeight - 1,VGAColors[clActiveCaption]);
      WriteSmOut(CL + 1,CT + (CaptionHeight Div 2) - 4,
                 Caption,VGAColors[clCaptionText] + VGAColors[clActiveCaption] Shl 8);
    End;

    { Paint children }

    I := 0;
    While I < Controls^.Count Do
    Begin
      PVGAControl(Controls^.At(I))^.Paint;
      Inc(I);
    End; { While }
    If Assigned(FOnPaint) Then FOnPaint(@Self,DummyEvent);
    VGAShowMouse;
  End;
End; { TVGAForm.Paint }

Procedure TVGAForm.FocusNextComponent;
Var
  I,J     : Integer;
  Found   : Boolean;
  Control : PVGAControl;

Begin
  If Not Updating Then
  Begin
    Updating := True;
    If FocusedComponent >= 0 Then
     PVGAControl(Controls^.At(FocusedComponent))^.SetFocused(False);
    Inc(TabNum);
    If TabNum >= MaxTab Then TabNum := 0;
    I     := Controls^.Count;
    Found := False;
    J     := FocusedComponent + 1;
    If J >= Controls^.Count Then J := 0;
    While (I > 0) And Not Found Do
    Begin
      Control := PVGAControl(Controls^.At(J));
      If Control^.TabOrder = TabNum Then
      Begin
        Control^.SetFocused(True);
        FocusedComponent := J;
        Found            := True;
      End
      Else
      Begin
        Dec(I);
        Inc(J);
        If J >= Controls^.Count Then J := 0;
      End;
    End; { While }
    Updating := False;
  End;
End; { TVGAForm.FocusNextComponent }

Procedure TVGAForm.SetVisibleControl(C: PVGAControl; V: Boolean);
Begin
  If Not Updating Then
  Begin
    Updating := True;
    If V Then
    Begin
      If Not C^.GetVisible Then
      Begin
        Paint;                       { Needs to be changed later }
      End;
    End
    Else
    Begin
      If C^.GetVisible Then
      Begin
        If C^.GetFocused Then
        Begin
          FocusNextComponent;
          C^.Paint;
        End;
      End;
    End;
    Updating := False;
  End;
End; { TVGAForm.SetVisibleControl }

Procedure TVGAForm.SetFocusedControl(C: PVGAControl; V: Boolean);
Begin
  If Not Updating Then
  Begin
    Updating := True;
    If Not V Then
    Begin
      If C^.GetFocused Then FocusNextComponent;
    End
    Else
    Begin
      If Not C^.GetFocused Then
      Begin
        If FocusedComponent >= 0 Then
         PVGAControl(Controls^.At(FocusedComponent))^.SetFocused(False);
        FocusedComponent := Controls^.IndexOf(C);
        C^.SetFocused(True);
      End;
    End;
    Updating := False;
  End;
End; { TVGAForm.SetFocusedControl }

Procedure TVGAForm.Show;
Begin
  If Assigned(FOnShow) Then FOnShow(@Self,DummyEvent);
  VGAApplication^.ShowForm(@Self);
End; { TVGAForm.Show }

Procedure TVGAForm.ShowModal;
Begin
  Modal := True;
  Show;
End; { TVGAForm.ShowModal }

Procedure TVGAForm.Close;
Begin
  VGAApplication^.CloseForm(@Self);
  If Assigned(FOnClose) Then FOnClose(@Self,DummyEvent);
End; { TVGAForm.Close }

Procedure TVGAForm.HandleEvent(Var Event: TEvent);
Var C: PVGAControl;
Begin
  Case Event.What Of
    evMouseUp,evMouseDown,evMouseMove,evMouseAuto:
    Begin
      C := GetControlAt(Event.Where.X - GetAbsLeft,Event.Where.Y - GetAbsTop,False);
      If C <> Nil Then
      Begin
        If (C^.TabOrder <> -1) And (Event.What = evMouseDown) Then SetFocusedControl(C,True);
        C^.HandleEvent(Event);
      End;
    End;
    evKeyDown:
    Begin
      Case Event.CharCode Of
      #9: FocusNextComponent
      Else
        If FocusedComponent >= 0 Then
         PVGAComponent(Controls^.At(FocusedComponent))^.HandleEvent(Event);
      End; { Case }
      If (Event.What = evKeyDown) And Assigned(Default) Then
       Default^.HandleEvent(Event);
    End;
  End; { Case }
End; { TVGAForm.HandleEvent }

Function TVGAForm.GetControlAt(X,Y: Integer; Any: Boolean): PVGAControl;
{ ------------------------------------------------------------------ }
{ Find the control at the X,Y position. X and Y are relative to the  }
{ client position of the form                                        }
{ ------------------------------------------------------------------ }
Var
  I     : Integer;
  Found : Boolean;
  C1,C2 : PVGAControl;

Begin
  I     := 0;
  Found := False;
  C2    := Nil;
  While (I < Controls^.Count) And Not Found Do
  Begin
    C1 := PVGAControl(Controls^.At(I));
    If C1^.InBounds(X,Y) Then
    Begin
      C2    := C1^.GetControlAt(X - C1^.Left,Y - C1^.Top,Any);
      Found := True;
    End
    Else Inc(I);
  End; { While }
  GetControlAt := C2;
End; { TVGAForm.GetControlAt }

{ ----------------------------------------- }
{ TVGAChooser                               }
{ ----------------------------------------- }

Procedure TVGAChooser_btnOkClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAChooser(Sender^.Owner)^.btnOkClick(Sender^.Parent,Event);
End; { TVGAChooser_btnOkClick }

Procedure TVGAChooser_btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAChooser(Sender^.Owner)^.btnCancelClick(Sender^.Parent,Event);
End; { TVGAChooser_btnCancelClick }

Procedure TVGAChooser_lbMainChange(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAChooser(Sender^.Owner)^.lbMainChange(Sender^.Parent,Event);
End; { TVGAChooser_lbMainDblClick }

Procedure TVGAChooser_lbMainDblClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAChooser(Sender^.Owner)^.lbMainDblClick(Sender^.Parent,Event);
End; { TVGAChooser_lbMainDblClick }

Procedure TVGAChooser.btnOkClick(Sender: PVGAControl; Var Event: TEvent);
Var
  St : String;
  I  : Integer;

Begin
  St := edtFile^.GetText;
  If (Pos(':',St) <> 0) Or
     (Copy(St,1,1) = '\')
   Then St := FExpand(St)
   Else St := FExpand(InitialDir + St);
  If ValidFileName(St) Then
  Begin
    Chose    := True;
    FileName := St;
    Close;
  End
  Else
  Begin
    If IsDirectory(St) Then
    Begin
      If Copy(St,Length(St),1) <> '\' Then St := St + '\';
      St := St + Filter;
    End;
    I := Length(St) + 1;
    Repeat
      Dec(I);
    Until (I = 0) Or (St[I] In [':','\']);
    If I < Length(St) Then Filter := Copy(St,I + 1,Length(St) - I);
    If I > 0 Then
    Begin
      St := Copy(St,1,I);
      If (Pos(':',St) <> 0) Or
         (Copy(St,1,1) = '\')
       Then InitialDir := FExpand(St)
       Else InitialDir := FExpand(InitialDir + St);
      If Copy(InitialDir,Length(InitialDir),1) <> '\' Then InitialDir := InitialDir + '\';
    End;
    edtFile^.SetText(Filter);
    LoadListBox;
    Paint;
  End;
End; { TVGAChooser.btnOkClick }

Procedure TVGAChooser.btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  Chose    := False;
  FileName := '';
  Close;
End; { TVGAChooser.btnCancelClick }

Procedure TVGAChooser.lbMainChange(Sender: PVGAControl; Var Event: TEvent);
Begin
  If (lbMain^.ItemIndex >= 0) And (lbMain^.ItemIndex < lbMain^.Items^.Count) Then
   edtFile^.SetText(StrPas(PStringObj(lbMain^.Items^.At(lbMain^.ItemIndex))^.P));
End; { TVGAChooser.lbMainChange }

Procedure TVGAChooser.lbMainDblClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  If ValidFileName(InitialDir + edtFile^.GetText) Then
  Begin
    Chose    := True;
    FileName := edtFile^.GetText;
    Close;
  End
  Else
  Begin
    InitialDir := FExpand(InitialDir + edtFile^.GetText);
    If Copy(InitialDir,Length(InitialDir),1) <> '\' Then InitialDir := InitialDir + '\';
    edtFile^.SetText(Filter);
    LoadListBox;
    Paint;
  End;
End; { TVGAChooser.lbMainDblClick }



Constructor TVGAChooser.Init(AEnabled: Boolean; Const ACaption,AFilter,AInitialDir: String);
Begin
  Inherited Init(10,10,ScreenWidth - 20,ScreenHeight - 20,AEnabled,
                 clSilver,ACaption,bsDialog);
  FileName   := '';
  Filter     := AFilter;
  If AInitialDir = ''
   Then GetDir(0,InitialDir)
   Else InitialDir := AInitialDir;
  If Copy(InitialDir,Length(InitialDir),1) <> '\' Then InitialDir := InitialDir + '\';
  lbMain      := New(PVGAListBox,Init(4,4,ClientWidth - 8,ClientHeight - 59,True,True,clWindow,orHorizontal,90,10));
  btnOk       := New(PVGAButton,Init(ClientWidth - 55,lbMain^.Top + lbMain^.Height + 4,51,17,True,True,clBtnFace,'Ok'));
  btnCancel   := New(PVGAButton,Init(ClientWidth - 55,lbMain^.Top + lbMain^.Height + 25,51,17,True,True,clBtnFace,'Cancel'));
  If ClientWidth >= 200 Then
  Begin
    edtFile     := New(PVGAEdit,Init(80,lbMain^.Top + lbMain^.Height + 4,ClientWidth - 160,15,True,True,clWindow,Filter));
    lblFileName := New(PVGALabel,Init(4,edtFile^.Top + 3,60,8,True,True,clBtnText,'File name:'));
    lblPath     := New(PVGALabel,Init(4,btnCancel^.Top + 4,edtFile^.Left + edtFile^.Width - 4,8,True,True,
                                      clBtnText,InitialDir + Filter));
  End
  Else
  Begin
    lblFileName := New(PVGALabel,Init(4,lbMain^.Top + lbMain^.Height + 4,64,8,True,True,clBtnText,'File name:'));
    edtFile     := New(PVGAEdit,Init(4,lblFileName^.Top + lblFileName^.Height + 4,64,15,True,True,clWindow,Filter));
    lblPath     := New(PVGALabel,Init(4,edtFile^.Top + edtFile^.Height + 4,64,8,True,True,
                                      clBtnText,InitialDir + Filter));
  End;
  AddControl(edtFile);
  AddControl(btnOk);
  AddControl(btnCancel);
  AddControl(lbMain);
  AddControl(lblFileName);
  AddControl(lblPath);

  btnOk^.AssignEvent(etOnClick,TVGAChooser_btnOkClick);
  btnCancel^.AssignEvent(etOnClick,TVGAChooser_btnCancelClick);
  lbMain^.AssignEvent(etOnChange,TVGAChooser_lbMainChange);
  lbMain^.AssignEvent(etOnDblClick,TVGAChooser_lbMainDblClick);
  Default := btnOk;
  Chose   := False;
  LoadListBox;
End; { TVGAChooser.Init }

Procedure TVGAChooser.LoadListBox;
Var
  S : SearchRec;
  F : PSortedStringCollection;
  I : Integer;

Begin
  lbMain^.Clear;
  lblPath^.SetText(InitialDir + Filter);
  F := New(PSortedStringCollection,Init(10,10));

  { Load files only }

  FindFirst(InitialDir + Filter,Archive + ReadOnly + Hidden + SysFile,S);
  While DosError = 0 Do
  Begin
    F^.Insert(New(PStringObj,Init(S.Name)));
    FindNext(S);
  End; { While }
  I := 0;
  While I < F^.Count Do
  Begin
    lbMain^.AddString(StrPas(PStringObj(F^.At(I))^.P));
    Inc(I);
  End; { While }
  F^.FreeAll;

  { Load directories only }

  FindFirst(InitialDir + '*.*',AnyFile,S);
  While DosError = 0 Do
  Begin
    If ((S.Attr And Directory) <> 0) And (S.Name <> '.') Then F^.Insert(New(PStringObj,Init(S.Name + '\')));
    FindNext(S);
  End; { While }
  I := 0;
  While I < F^.Count Do
  Begin
    lbMain^.AddString(StrPas(PStringObj(F^.At(I))^.P));
    Inc(I);
  End; { While }
  Dispose(F,Done);
End; { TVGAChooser.LoadListBox }

Procedure TVGAChooser.SetFilter(Const AFilter: String);
Begin
  Filter := AFilter;
  LoadListBox;
  Paint;
End; { TVGAChooser.SetFilter }

Function TVGAChooser.GetFilter: String;
Begin
  GetFilter := Filter;
End; { TVGAChooser.GetFilter }

Procedure TVGAChooser.SetInitialDir(Const AInitialDir: String);
Begin
  InitialDir := AInitialDir;
  LoadListBox;
  Paint;
End; { TVGAChooser.SetInitialDir }

Function TVGAChooser.GetInitialDir: String;
Begin
  GetInitialDir := InitialDir;
End; { TVGAChooser.GetInitialDir }

Destructor TVGAChooser.Done;
Begin
  Inherited Done;
End; { TVGAChooser.Done }

Procedure TVGAChooser.HandleEvent(Var Event: TEvent);
Begin
  Case Event.What Of
    evKeyDown:
    Begin
      If Event.CharCode = #27 Then
      Begin
        Chose    := False;
        FileName := '';
        Close;
        Event.What := evNothing;  { Dispatch event }
      End;
    End;
  End; { Case }
  Inherited HandleEvent(Event);
End; { TVGAChooser.HandleEvent }

{ ----------------------------------------- }
{ TVGAMessageBox                            }
{ ----------------------------------------- }

Procedure TVGAMessageBox_btnOkClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAMessageBox(Sender^.Owner)^.btnOkClick(Sender^.Parent,Event);
End; { TVGAMessageBox_btnOkClick }

Procedure TVGAMessageBox_btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAMessageBox(Sender^.Owner)^.btnCancelClick(Sender^.Parent,Event);
End; { TVGAMessageBox_btnCancelClick }

Procedure TVGAMessageBox.btnOkClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  Result := idOK;
  Close;
End; { TVGAMessageBox.btnOkClick }

Procedure TVGAMessageBox.btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  Result := idCancel;
  Close;
End; { TVGAMessageBox.btnCancelClick }

Constructor TVGAMessageBox.Init(AEnabled: Boolean; Const ACaption,AMessage: String; Buttons: TMsgBoxButton);
Var X: Integer;
Begin
  Inherited Init(20,(ScreenHeight Div 2) - 30,ScreenWidth - 40,60,AEnabled,
                 clSilver,ACaption,bsDialog);
  If (mbOk In Buttons) And (mbCancel In Buttons)
   Then X := (ClientWidth Div 2) - 53
   Else X := (ClientWidth Div 2) - 25;
  If mbOk In Buttons Then
  Begin
    btnOk := New(PVGAButton,Init(X,ClientHeight - 30,51,17,True,True,clBtnFace,'Ok'));
    btnOk^.AssignEvent(etOnClick,TVGAMessageBox_btnOkClick);
    AddControl(btnOk);
    Inc(X,55);
  End;
  If mbCancel In Buttons Then
  Begin
    btnCancel := New(PVGAButton,Init(X,ClientHeight - 30,51,17,True,True,clBtnFace,'Cancel'));
    btnCancel^.AssignEvent(etOnClick,TVGAMessageBox_btnCancelClick);
    AddControl(btnCancel);
    Inc(X,55);
  End;
  lblMessage := New(PVGALabel,Init(4,4,ClientWidth - 8,8,True,True,clBtnText,AMessage));
  AddControl(lblMessage);
End; { TVGAMessageBox.Init }

Destructor TVGAMessageBox.Done;
Begin
  Inherited Done;
End; { TVGAMessageBox.Done }

Procedure TVGAMessageBox.HandleEvent(Var Event: TEvent);
Begin
  Case Event.What Of
    evKeyDown:
    Begin
      If Event.CharCode = #27 Then
      Begin
        Close;
        Event.What := evNothing;  { Dispatch event }
      End;
    End;
  End; { Case }
  Inherited HandleEvent(Event);
End; { TVGAMessageBox.HandleEvent }

{ ----------------------------------------- }
{ TVGAInputBox                              }
{ ----------------------------------------- }

Procedure TVGAInputBox_btnOkClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAInputBox(Sender^.Owner)^.btnOkClick(Sender^.Parent,Event);
End; { TVGAInputBox_btnOkClick }

Procedure TVGAInputBox_btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  PVGAInputBox(Sender^.Owner)^.btnCancelClick(Sender^.Parent,Event);
End; { TVGAInputBox_btnCancelClick }

Procedure TVGAInputBox.btnOkClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  Text := edtInput^.GetText;
  Close;
End; { TVGAInputBox.btnOkClick }

Procedure TVGAInputBox.btnCancelClick(Sender: PVGAControl; Var Event: TEvent);
Begin
  Text := InitialText;
  Close;
End; { TVGAInputBox.btnCancelClick }

Constructor TVGAInputBox.Init(AEnabled: Boolean; Const ACaption,AMessage,AInitialText: String);
Begin
  Inherited Init(20,(ScreenHeight Div 2) - 37,ScreenWidth - 40,74,AEnabled,
                 clSilver,ACaption,bsDialog);
  InitialText := AInitialText;
  btnOk       := New(PVGAButton,Init((ClientWidth Div 2) - 55,ClientHeight - 30,51,17,True,True,clBtnFace,'Ok'));
  btnCancel   := New(PVGAButton,Init((ClientWidth Div 2) + 2,ClientHeight - 30,51,17,True,True,clBtnFace,'Cancel'));
  edtInput    := New(PVGAEdit,Init(4,16,ClientWidth - 8,15,True,True,clWindow,InitialText));
  lblMessage  := New(PVGALabel,Init(4,4,ClientWidth - 8,8,True,True,clBtnText,AMessage));

  AddControl(edtInput);
  AddControl(btnOk);
  AddControl(btnCancel);
  AddControl(lblMessage);

  btnOk^.AssignEvent(etOnClick,TVGAInputBox_btnOkClick);
  btnCancel^.AssignEvent(etOnClick,TVGAInputBox_btnCancelClick);
  Default := btnOk;
End; { TVGAInputBox.Init }

Destructor TVGAInputBox.Done;
Begin
  Inherited Done;
End; { TVGAInputBox.Done }

Function TVGAInputBox.GetText: String;
Begin
  GetText := Text;
End; { TVGAInputBox.GetText }

Procedure TVGAInputBox.HandleEvent(Var Event: TEvent);
Begin
  Case Event.What Of
    evKeyDown:
    Begin
      If Event.CharCode = #27 Then
      Begin
        Text := InitialText;
        Close;
        Event.What := evNothing;  { Dispatch event }
      End;
    End;
  End; { Case }
  Inherited HandleEvent(Event);
End; { TVGAInputBox.HandleEvent }

Begin
  ScreenWidth  := 320;
  ScreenHeight := 200;
  ModifyCode;
End.

// ----------------------------------------------------------------------
// PCAE - PC Atari Emulator - Atari 2600 Emulator for MS-DOS
// Copyright (C) 2000 John Dullea
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// ----------------------------------------------------------------------

